﻿using UnityEngine;
using UnityEditor;
using System.Diagnostics;
using UnityEditor.SceneManagement;

namespace Igloo.Common
{
    [CustomEditor(typeof(IglooManager))]
    public class IglooManagerEditor : Editor
    {
        #region Serialized Properties
        SerializedProperty createOnAwake;
        SerializedProperty dontDestroyOnLoad;
        SerializedProperty customCameraPrefab;
        SerializedProperty followObject;
        SerializedProperty spoutResources;
        SerializedProperty ndiResources;
#if IGLOO_UI
        SerializedProperty canvasUI;
#endif

        Texture logo;
        IglooManager manager;
        #endregion

        /// <summary>
        /// Adds the Igloo Manager gameobject to the current scene at 0,0,0
        /// </summary>
        [MenuItem("Tools/Igloo/Spawn Igloo Manager")]
        public static void SpawnIglooManager()
        {
            if (FindObjectOfType<IglooManager>()) return;
            GameObject managerObj = GameObject.Instantiate(AssetDatabase.LoadAssetAtPath<GameObject>("Packages/com.igloo.toolkit/Runtime/Resources/Prefabs/IglooManager.prefab")) as GameObject;
            if (managerObj == null)
            {
                UnityEngine.Debug.LogError("<b>[Igloo]</b> Could not locate Igloo Manager in Igloo Toolkit Resources it must be added manually.");
                return; }
            StageUtility.PlaceGameObjectInCurrentStage(managerObj); 
            managerObj.name = "IglooManager";
#if KLAK_SPOUT
            managerObj.GetComponent<IglooManager>().spoutResources = AssetDatabase.LoadAssetAtPath<Klak.Spout.SpoutResources>("Packages/jp.keijiro.klak.spout/Editor/SpoutResources.asset");
#endif
#if KLAK_NDI
            managerObj.GetComponent<IglooManager>().ndiResources = AssetDatabase.LoadAssetAtPath<Klak.Ndi.NdiResources>("Packages/jp.keijiro.klak.ndi/Runtime/Resource/NDIResources.asset");
#endif
        }

        private void OnEnable()
        {
            createOnAwake = serializedObject.FindProperty("createOnAwake");
            dontDestroyOnLoad = serializedObject.FindProperty("dontDestroyOnLoad");
            customCameraPrefab = serializedObject.FindProperty("customCameraPrefab");
            followObject = serializedObject.FindProperty("followObject");
            spoutResources = serializedObject.FindProperty("spoutResources");
            ndiResources = serializedObject.FindProperty("ndiResources");
#if IGLOO_UI
            canvasUI = serializedObject.FindProperty("canvasUI");
#endif
        }


        /// <summary>
        /// Changes the view of the Igloo Manager inspector, depending on Regular or Pro Unity.
        /// Also adds the open settings, and add spout resources button to the package. (Klak Depending)
        /// </summary>
        public override void OnInspectorGUI()
        {

            var style = new GUIStyle(GUI.skin.label)
            {
                fixedHeight = 25f,
                alignment = TextAnchor.MiddleCenter,
                fontStyle = FontStyle.Bold,
                fontSize = 20,
            };
            var style2 = new GUIStyle(GUI.skin.label)
            {
                alignment = TextAnchor.MiddleCenter,
                fontStyle = FontStyle.Bold,
                fontSize = 12,
            };
            logo = UnityEditorInternal.InternalEditorUtility.HasPro() ? Resources.Load<Texture>("Textures/IglooLogoWhite") : Resources.Load<Texture>("Textures/IglooLogoBlack");
            manager = (IglooManager)target;
            serializedObject.Update();

            EditorGUILayout.BeginVertical("GroupBox");
            if (logo)
            {
                GUILayout.Box(logo, GUILayout.Width(200), GUILayout.Height(108), GUILayout.ExpandWidth(true));
                EditorGUILayout.Separator();
            }

            EditorGUILayout.BeginVertical("textArea", GUILayout.Height(30));
            EditorGUILayout.LabelField("IGLOO MANAGER", style, GUILayout.ExpandWidth(true));
            EditorGUILayout.EndVertical();
            EditorGUILayout.EndVertical();

            GUILayout.BeginVertical("", "GroupBox");
            EditorGUILayout.PropertyField(createOnAwake);
            EditorGUILayout.PropertyField(dontDestroyOnLoad);
            EditorGUILayout.Separator();
            EditorGUILayout.PropertyField(customCameraPrefab);
            EditorGUILayout.PropertyField(followObject);
            EditorGUILayout.Separator();
#if IGLOO_UI
            EditorGUILayout.PropertyField(canvasUI);
            EditorGUILayout.Separator();
#endif
#if KLAK_SPOUT
            if(manager.spoutResources != null) EditorGUILayout.PropertyField(spoutResources);
#endif
#if KLAK_NDI
            if(manager.ndiResources != null) EditorGUILayout.PropertyField(ndiResources);
#endif
            EditorGUILayout.EndVertical();

            #region Spout and NDI
#if KLAK_SPOUT && KLAK_NDI
            if (manager.spoutResources == null || manager.ndiResources == null)
            {
                GUILayout.BeginVertical("", "GroupBox");
                EditorGUILayout.BeginVertical("textArea", GUILayout.Height(20));
                EditorGUILayout.LabelField("Add Resources", style2, GUILayout.ExpandWidth(true));
                EditorGUILayout.EndVertical();
            }
#endif
#if KLAK_SPOUT
            if (manager.spoutResources == null)
            {
                EditorGUILayout.Separator();
                if (GUILayout.Button("Add Spout Resources", GUILayout.Height(20)))
                {
                    FindAndAddSpoutResources();
                    EditorUtility.SetDirty(manager);
                }
            }
#endif
#if KLAK_NDI
            if (manager.ndiResources == null)
            {
                EditorGUILayout.Separator();
                if (GUILayout.Button("Add NDI Resources", GUILayout.Height(20)))
                {
                    FindAndAddNDIResources();
                    EditorUtility.SetDirty(manager);
                }
            }
#endif
#if KLAK_SPOUT && KLAK_NDI
            if (manager.spoutResources == null || manager.ndiResources == null)
            {
                EditorGUILayout.EndVertical();
            }
#endif
#endregion

            GUILayout.BeginVertical("", "GroupBox");
            if (GUILayout.Button("Open Settings Wizard", GUILayout.Height(20)))
            {
                SettingsWizard.Init();
            }
            EditorGUILayout.Separator();
            if (GUILayout.Button("Open Settings File", GUILayout.Height(20)))
            {
                string path = System.IO.Path.Combine(Utils.GetDataPath(), "IglooSettings.xml");
                Process.Start(path);
            }
            EditorGUILayout.EndVertical();
            serializedObject.ApplyModifiedProperties();
        }

        /// <summary>
        /// Looks through the packages for the Klak Spout package, and adds the SpoutResources file to
        /// the Igloo Manager system. As this file is located in the packages section, it's hard to find. 
        /// This makes it easier for users.
        /// </summary>
        private void FindAndAddSpoutResources() 
        {
#if KLAK_SPOUT
            
            manager.spoutResources = AssetDatabase.LoadAssetAtPath<Klak.Spout.SpoutResources>("Packages/jp.keijiro.klak.spout/Editor/SpoutResources.asset");
            if(manager.spoutResources == null) {
                UnityEngine.Debug.LogWarning("<b>[Igloo]</b> Could not locate Spout Resources in the Klak Spout package folder, it must be added manually.");
            } 
#endif
        }

        /// <summary>
        /// Looks through the packages for the Klak NDI package, and adds the NDIResources file to
        /// the Igloo Manager system. As this file is located in the packages section, it's hard to find. 
        /// This makes it easier for users.
        /// </summary>
        private void FindAndAddNDIResources() {
#if KLAK_NDI

            manager.ndiResources = AssetDatabase.LoadAssetAtPath<Klak.Ndi.NdiResources>("Packages/jp.keijiro.klak.ndi/Runtime/Resource/NDIResources.asset");
            if (manager.spoutResources == null) {
                UnityEngine.Debug.LogWarning("<b>[Igloo]</b> Could not locate NDI Resources in the Klak NDI package folder, it must be added manually.");
            }
#endif
        }
    }
}

