﻿using UnityEngine;
using UnityEditor;
using Igloo.Controllers;
using System.IO;
using Unity.Plastic.Newtonsoft.Json.Linq;
using System.Diagnostics;

namespace Igloo.Common
{
    public class SettingsWizard : EditorWindow
    {
        Texture logo;

        Settings settings;
        Settings cachedSettings;
        Vector2 scrollPos;

        [MenuItem("Tools/Igloo/Settings Wizard")]
        public static void Init()
        {
            SettingsWizard window = (SettingsWizard)EditorWindow.GetWindow(typeof(SettingsWizard), true, "Igloo Settings Wizard", true);
            window.Show();
        }

        public void LoadSettings()
        {
            settings = File.Exists(Utils.GetDataPath() + "/IglooSettings.xml") ? LoadSettingsFromStreamingAssets() : LoadDefaultSettings();
            cachedSettings = settings;
        }

        public Settings LoadSettingsFromStreamingAssets()
        {
            Serializer serializer = new Serializer();
            Settings settings = serializer.Load(System.IO.Path.Combine(Utils.GetDataPath() + "/", "IglooSettings"));
            UnityEngine.Debug.Log("<b>[Igloo]</b> Found Local Settings, and loaded.");
            return settings;
        }

        public Settings LoadDefaultSettings()
        {
            UnityEngine.Debug.Log("<b>[Igloo]</b> No local settings, loaded defaults.");
            return DefaultConfigurations.Shipping();
        }

        public bool SaveSettings(Settings _settings, bool saveAs = false)
        {
            settings.version = Utils.GetVersion();
            Igloo.Serializer serializer = new Serializer();
            string path = System.IO.Path.Combine(Utils.GetDataPath() + "/", "IglooSettings");
            if (saveAs) path = EditorUtility.SaveFilePanel("Set Save Location", "", "IglooSettings", "xml");
            serializer.Save(path, settings);
            return true;
        }

        private void OnGUI()
        {
            var style = new GUIStyle(GUI.skin.label)
            {
                fixedHeight = 25f,
                alignment = TextAnchor.MiddleCenter,
                fontStyle = FontStyle.Bold,
                fontSize = 20,
            };
            var style2 = new GUIStyle(GUI.skin.label)
            {
                alignment = TextAnchor.MiddleCenter,
                fontStyle = FontStyle.Bold,
                fontSize = 12,
            };

            if (settings == null) LoadSettings();
            EditorGUILayout.BeginVertical("GroupBox");
            scrollPos = GUILayout.BeginScrollView(scrollPos, false, true, GUILayout.ExpandHeight(true));
            logo = UnityEditorInternal.InternalEditorUtility.HasPro() ? Resources.Load<Texture>("Textures/IglooLogoWhite") : Resources.Load<Texture>("Textures/IglooLogoBlack");
            if (logo)
            {
                GUILayout.Box(logo, GUILayout.Width(200), GUILayout.Height(108), GUILayout.ExpandWidth(true));
                EditorGUILayout.Separator();
            }

            if (settings != null)
            {

#if IGLOO_PLAYER
                EditorGUILayout.BeginVertical("textArea", GUILayout.Height(30));
                EditorGUILayout.LabelField("PLAYER SETTINGS", style, GUILayout.ExpandWidth(true));
                EditorGUILayout.EndVertical();

                GUILayout.BeginVertical("", "GroupBox");
                GUILayout.Box("If you have installed the Igloo Player, you can enable it here and adjust how your controller influences it.");
                EditorGUILayout.Separator();
                settings.PlayerSettings.usePlayer = EditorGUILayout.Toggle("Use Player", settings.PlayerSettings.usePlayer);
                if (settings.PlayerSettings.usePlayer)
                {
                    if (settings.HeadSettings.headtracking)
                    {
                        settings.PlayerSettings.optitrackControllerRigidBodyID = EditorGUILayout.IntField("Controller Rigidbody ID", settings.PlayerSettings.optitrackControllerRigidBodyID);
                        settings.HeadSettings.optitrackServerIP = EditorGUILayout.TextField("Optitrack Server IP:", settings.HeadSettings.optitrackServerIP);
                        GUILayout.Space(10f);
                    }
                    settings.PlayerSettings.rotationMode = (int)(PlayerManager.ROTATION_MODE)EditorGUILayout.EnumPopup("Rotation Mode:", (PlayerManager.ROTATION_MODE)settings.PlayerSettings.rotationMode);
                    settings.PlayerSettings.movementInput = (int)(PlayerManager.MOVEMENT_INPUT)EditorGUILayout.EnumPopup("Movement Input:", (PlayerManager.MOVEMENT_INPUT)settings.PlayerSettings.movementInput);
                    settings.PlayerSettings.movementMode = (int)(PlayerManager.MOVEMENT_MODE)EditorGUILayout.EnumPopup("Movement Mode:", (PlayerManager.MOVEMENT_MODE)settings.PlayerSettings.movementMode);
                    settings.PlayerSettings.walkSpeed = EditorGUILayout.FloatField("Walk Speed:", settings.PlayerSettings.walkSpeed);
                    settings.PlayerSettings.runSpeed = EditorGUILayout.FloatField("Run Speed:", settings.PlayerSettings.runSpeed);
                    settings.PlayerSettings.smoothTime = EditorGUILayout.FloatField("Rotation Smoothing:", settings.PlayerSettings.smoothTime);
                    settings.PlayerSettings.crosshairHideMode = (int)(Crosshair.CROSSHAIR_MODE)EditorGUILayout.EnumPopup("Crosshair Hide Mode:", (Crosshair.CROSSHAIR_MODE)settings.PlayerSettings.crosshairHideMode);
                }
                EditorGUILayout.EndVertical();
#endif

                EditorGUILayout.BeginVertical("textArea", GUILayout.Height(30));
                EditorGUILayout.LabelField("NETWORK SETTINGS", style, GUILayout.ExpandWidth(true));
                EditorGUILayout.EndVertical();

                GUILayout.BeginVertical("", "GroupBox");
                GUILayout.Box("These network settings govern how the application interacts with the ICE communication system.");
                EditorGUILayout.Separator();
                settings.NetworkSettings.inPort = EditorGUILayout.IntField("In Port:", settings.NetworkSettings.inPort);
                settings.NetworkSettings.outPort = EditorGUILayout.IntField("Out Port:", settings.NetworkSettings.outPort);

                char separator = '.';
                string[] ip = settings.NetworkSettings.outIP.Split(separator);

                if (ip.Length == 4)
                {
                    GUILayout.BeginHorizontal();
                    GUILayout.Label("IP:");
                    ip[0] = EditorGUILayout.IntField("", int.Parse(ip[0]), GUILayout.MaxWidth(100.0f)).ToString();
                    ip[1] = EditorGUILayout.IntField("", int.Parse(ip[1]), GUILayout.MaxWidth(100.0f)).ToString();
                    ip[2] = EditorGUILayout.IntField("", int.Parse(ip[2]), GUILayout.MaxWidth(100.0f)).ToString();
                    ip[3] = EditorGUILayout.IntField("", int.Parse(ip[3]), GUILayout.MaxWidth(100.0f)).ToString();
                    GUILayout.EndHorizontal();
                    settings.NetworkSettings.outIP = ip[0] + "." + ip[1] + "." + ip[2] + "." + ip[3];
                }
                EditorGUILayout.EndVertical();


                EditorGUILayout.BeginVertical("textArea", GUILayout.Height(30));
                EditorGUILayout.LabelField("HEADTRACKING SETTINGS", style, GUILayout.ExpandWidth(true));
                EditorGUILayout.EndVertical();

                GUILayout.BeginVertical("", "GroupBox");
                GUILayout.Box("Leverage Optitrack or VRPN tracking data to enhance the Igloo viewing experience");
                EditorGUILayout.Separator();
                settings.HeadSettings.headtracking = EditorGUILayout.Toggle("Use Headtracking:", settings.HeadSettings.headtracking);
                if (settings.HeadSettings.headtracking) {
                    settings.HeadSettings.headTrackingInput = (int)(HeadManager.HeadTrackingInput)EditorGUILayout.EnumPopup("HeadTracking Input:", (HeadManager.HeadTrackingInput)settings.HeadSettings.headTrackingInput);
                    if (settings.HeadSettings.headTrackingInput == 1) {
                        settings.HeadSettings.optitrackServerIP = EditorGUILayout.TextField("Optitrack Server IP:", settings.HeadSettings.optitrackServerIP);
                        settings.HeadSettings.optitrackHeadRigidBodyID = EditorGUILayout.IntField("Head Rigidbody ID:", settings.HeadSettings.optitrackHeadRigidBodyID);

                    }
                }
                EditorGUILayout.EndVertical();


#if IGLOO_UI
                EditorGUILayout.BeginVertical("textArea", GUILayout.Height(30));
                EditorGUILayout.LabelField("UI SETTINGS", style, GUILayout.ExpandWidth(true));
                EditorGUILayout.EndVertical();

                GUILayout.BeginVertical("", "GroupBox");
                GUILayout.Box("If you have the Igloo world space UI installed, you can enable it here, and adjust how the player interacts with it.");
                EditorGUILayout.Separator();
                settings.UISettings.useUI = EditorGUILayout.Toggle("Use UI", settings.UISettings.useUI);
                if (settings.UISettings.useUI)
                {
                    EditorGUILayout.HelpBox("Screen name must be either 'Cylinder' or 'Plane' , unless using custom geometry", MessageType.Info);
                    settings.UISettings.screenName = EditorGUILayout.TextField("Screen Name:", settings.UISettings.screenName);
                    settings.UISettings.screenPos = new Vector3Item(EditorGUILayout.Vector3Field("Screen Position:", settings.UISettings.screenPos.Vector3));
                    settings.UISettings.screenRot = new Vector3Item(EditorGUILayout.Vector3Field("Screen Rotation:", settings.UISettings.screenRot.Vector3));
                    settings.UISettings.screenScale = new Vector3Item(EditorGUILayout.Vector3Field("Screen Scale:", settings.UISettings.screenScale.Vector3));
                    settings.UISettings.followCrosshair = EditorGUILayout.Toggle("Follow Crosshair:", settings.UISettings.followCrosshair);
                    settings.UISettings.followSpeed = EditorGUILayout.FloatField("Follow Speed:", settings.UISettings.followSpeed);
                    settings.UISettings.debugUIMode = EditorGUILayout.Toggle("Debug UI Cursor", settings.UISettings.debugUIMode);
                    settings.TouchScreenSettings.UseTouchScreen = EditorGUILayout.Toggle("Use Touch Screen System", settings.TouchScreenSettings.UseTouchScreen);
                    if (settings.TouchScreenSettings.UseTouchScreen) {
                        settings.TouchScreenSettings.XPositionStart = EditorGUILayout.FloatField("Window Start Horizontal Value", settings.TouchScreenSettings.XPositionStart);
                        settings.TouchScreenSettings.XPositionEnd = EditorGUILayout.FloatField("Window End Horizontal Value", settings.TouchScreenSettings.XPositionEnd);
                        settings.TouchScreenSettings.YPositionStart = EditorGUILayout.FloatField("Window Start Vertical Value", settings.TouchScreenSettings.YPositionStart);
                        settings.TouchScreenSettings.YPositionEnd = EditorGUILayout.FloatField("Window End Vertical Value", settings.TouchScreenSettings.YPositionEnd);
                    }
                }
                EditorGUILayout.EndVertical();

#endif
            }

            GUILayout.Box("Settings are saved when this window is closed", GUILayout.ExpandWidth(true));

            EditorGUILayout.EndScrollView();
            EditorGUILayout.EndVertical();

        }

        private void OnDestroy()
        {
            SaveSettings(settings);
        }
    }
}

