Shader "Igloo/TrueDisk"
{
    Properties
    {
        _MainTex("Texture", 2D) = "white" {}
        _u_width("Screen Width", Float) = 1280
        _u_height("Screen Height", Float) = 720

        _u_rotate_pitch_direction("Rotate Pitch Direction", Range(0, 360)) = 0
        _u_rotate_yaw("Rotate Yaw", Range(0, 360)) = 0
        _u_rotate_pitch("Rotate Pitch", Range(-90, 90)) = 0
        _u_rotate_pitch_thickness("Rotate Pitch Thickness", Range(0, 0.5)) = 0.1

        _u_disk_pa("Disk Point A", Vector) = (0,0,0)
        _u_disk_pb("Disk Point B", Vector) = (0,0,1)
        _u_disk_pc("Disk Point C", Vector) = (1,0,0)
        _u_disk_radius("Disk Radius", Range(0.01, 10.0)) = 1.0
        _u_disk_theta_offset("Disk Theta Offset", Range(0.0, 6.283185)) = 0.0
        _u_eye("Camera Position", Vector) = (0,0,0)
    }

    SubShader
    {
        // No culling or depth
        Cull Off ZWrite On ZTest LEqual

        Pass
        {
            CGPROGRAM
            #pragma vertex vert
            #pragma fragment frag

            #include "UnityCG.cginc"
            #define PI          3.1415926535897932
            #define PI_INV      0.3183098861837907

            struct appdata
            {
                float4 vertex : POSITION;
                float2 uv : TEXCOORD0;
            };

            struct v2f
            {
                float4 vertex : SV_POSITION;
                float2 uv : TEXCOORD0;
            };

            sampler2D _MainTex;

            // Texture uniforms
            uniform float _u_width;                     // screen width
            uniform float _u_height;                    // screen height

            // Rotation uniforms
            uniform float _u_rotate_yaw;                 // yaw rotation angle
            uniform float _u_rotate_pitch;               // pitch rotation angle
            uniform float _u_rotate_pitch_direction;     // pitch rotation plane angle
            uniform float _u_rotate_pitch_thickness;     // plane threshold width
            uniform float4x4 _u_rotate;                  // rotation matrix

            // Raytracer uniforms
            uniform float3 _u_disk_pa;
            uniform float3 _u_disk_pb;
            uniform float3 _u_disk_pc;
            uniform float _u_disk_radius;
            uniform float _u_disk_theta_offset;
            uniform float3 _u_eye;

            // In/Out variables
            float4 v_position : SV_POSITION;
            float2 v_texcoord : TEXCOORD0;
            float4 frag_color : SV_TARGET;

            // --------------------------------------------------------------------------------------
            // Convert from spherical (theta,phi)-coordinates to uv-coordinates.
            //
            float2 spherical_to_texcoord(float2 sphere_coord)
            {
                float phi = sphere_coord.x;
                float theta = sphere_coord.y;
                return float2(0.5 * (1.0 + phi * PI_INV), (theta * PI_INV));
            }

            //
            // Convert from uv-coordinates to spherical (theta,phi)-coordinates.
            //
            float2 texcoord_to_spherical(float2 texcoord)
            {
                float phi = PI * (2.0 * texcoord.x - 1.0);
                float theta = PI * texcoord.y;
                return float2(phi, theta);
            }

            //
            // Convert from cartesian to uv-coordinates coordinates.
            //
            float3 texcoord_to_cartesian(float2 coord)
            {
                float2 sphere_coord = texcoord_to_spherical(coord);
                float phi = sphere_coord.x;
                float theta = sphere_coord.y;
                return float3(
                    sin(theta) * sin(phi),
                    cos(theta),
                    sin(theta) * cos(phi));
            }

            //
            // Convert from uv-coordinates to cartesian coordinates.
            //
            float2 cartesian_to_texcoord(float3 position)
            {
                float theta = PI - acos(position.y);
                float phi = PI + atan2(position.x, position.z);
                return spherical_to_texcoord(float2(phi, theta));
            }

            // --------------------------------------------------------------------------------------
            // Compute the coordinates of a given pixel wrt to disk local frame of reference, defined
            // by the vertex positions pa, pb, and pc.
            //
            float3 pixel_coord(float2 pixel)
            {
                // Compute uv-coordinates of the given pixel.
                float2 uv = float2(1.0 - pixel.x, 1.0 - pixel.y);

                // Pixel position wrt to disk local reference frame
                float3 pos = uv.x * (_u_disk_pb - _u_disk_pa) + uv.y * (_u_disk_pc - _u_disk_pa);

                // Discard if the distance from the centre of the disk is outside the radius range or
                // if the polar angle is outside the rangle range.
                float3 ctr = 0.5 * (_u_disk_pb - _u_disk_pa) + 0.5 * (_u_disk_pc - _u_disk_pa);

                float radius = length(pos - ctr);
                if (radius > _u_disk_radius) {
                    discard;
                    // return float3(0.0, 0.0, 0.0);
                }

                float c = cos(_u_disk_theta_offset);
                float s = sin(_u_disk_theta_offset);
                float rx =  c * (uv.x - 0.5) + s * (uv.y - 0.5);
                float ry = -s * (uv.x - 0.5) + c * (uv.y - 0.5);
                uv.x = rx + 0.5;
                uv.y = ry + 0.5;

                // Return the pixel coordinates in world space.
                return ((1.0 - uv.x - uv.y) * _u_disk_pa) + (uv.x * _u_disk_pb) + (uv.y * _u_disk_pc);
            }

            // --------------------------------------------------------------------------------------
            // Ray-sphere intersection function.
            //
            bool intersect(in float3 ray_orig,
                in float3 ray_dir,
                in float3 sphere_centre,
                in float sphere_radius,
                out float hit_lambda)
            {
                // Solve the quadratic equation for the ray-sphere intersection.
                float3 orig = ray_orig - sphere_centre;   // ray origin relative to the sphere
                float a = dot(ray_dir, ray_dir);
                float b = 2.0 * dot(orig, ray_dir);
                float c = dot(orig, orig) - sphere_radius * sphere_radius;
                float d = b * b - 4.0 * a * c;

                // The solutions of the quadratic equation are given by t0 and t1.
                // Order t0 and t1 in increasing order:
                //  if t0 < 0 and t1 < 0,
                //      the object is behind the ray origin, there is no intersection.
                //  if t0 < 0,
                //      t1 > 0 is the intersection point.
                float t0, t1, tmp;

                if (d < 0.0 || a == 0.0) {      // d < 0, no intersection
                    return false;               // a = 0, ray dir is a zero vector
                }
                else if (d == 0.0) {
                    t0 = t1 = -0.5 * b / a;
                }
                else {
                    t0 = -0.5 * (b + sqrt(d)) / a;
                    t1 = c / (a * t0);
                }

                if (t0 > t1) {
                    tmp = t0;
                    t0 = t1;
                    t1 = tmp;
                }

                if (t0 < 0.0 && t1 < 0.0) {
                    return false;
                }
                else if (t0 < 0.0) {
                    t0 = t1;
                }

                // Compute the hit lambda line parameter.
                hit_lambda = t0;
                return true;
            }

            // --------------------------------------------------------------------------------------
            // Compute the intersection between a ray starting at the specified origin and along the
            // specified direction with a sphere with a give radius and centred at the given centre.
            //
            float3 raytrace(float3 rayOrig, float3 rayDir, float3 sphereCentre, float sphereRadius)
            {
                float hitLambda = 0.0;
                if (intersect(rayOrig, rayDir, sphereCentre, sphereRadius, hitLambda)) {
                    // Return hit position relative to the object position
                    float3 hitPos = rayOrig + hitLambda * rayDir - sphereCentre;
                    return normalize(hitPos);
                } else {
                    // discard;
                    return float3(0.0, 0.0, 0.0);
                }
            }

            // --------------------------------------------------------------------------------------
            // Sample the texture by ray-tracing the sphere.
            //
            float2 compute_texcoord(in float2 pixel)
            {
                // Compute the camera off-axis projection matrix. We are solving the inverse problem
                // here. We are given the pixel in screen space and want to find the direction given
                // the off axis projection matrix.
                float3 coord = pixel_coord(pixel);
                coord += float3(_u_eye.x, -_u_eye.y, _u_eye.z);

                float3 ray_orig = float3(0.0, 0.0, 0.0);        // ray origin
                float3 ray_dir = coord.xyz - ray_orig;          // ray direction
                normalize(ray_dir);

                // Initialize unit sphere
                float3 sphere_centre = float3(0.0, 0.0, 0.0);   // sphere position
                float sphere_radius = 1.0;                      // sphere radius

                // Cast a ray through this fragment. Ensure return hit position is normalized to the
                // sphere radius, unity in this case.
                float3 hitpos = raytrace(ray_orig, ray_dir, sphere_centre, sphere_radius);

                // Return the texture coordinates of the intersected point. position. Here, y-axis is
                // the up direction vector in a righ-handed coordinate system. Theta is the elevation
                // angle between the intersection vector and the the y-axis. Phi is the azimuth angle
                // in the xz-plane.
                return cartesian_to_texcoord(hitpos);
            }

            // --------------------------------------------------------------------------------------
            v2f vert(appdata v)
            {
                v2f o;
                o.vertex = UnityObjectToClipPos(v.vertex);
                o.uv = v.uv;
                return o;
            }

            //fixed4 frag(v2f i) : SV_Target
            //{
            //    fixed4 col = tex2D(_MainTex, i.uv);
            //    return col;
            //}

            fixed4 frag(v2f i) : SV_Target
            {
                // Compute the texture coordinates of the sphere interesection point.
                float2 texel = compute_texcoord(i.uv);
                // return fixed4(texel, 0.0, 1.0);

                // Sample the texture
                fixed4 col = tex2D(_MainTex, texel);
                return col;
            }

            ENDCG
        }
    }
}