﻿using System.Collections.Generic;
using UnityEngine;
using System.IO;
using Igloo.Common;
using System;
using System.Linq;
#if KLAK_SPOUT
using Klak.Spout;
#endif
#if KLAK_NDI
using Klak.Ndi;
#endif

namespace Igloo
{
#pragma warning disable IDE0090 // Use New()...
#pragma warning disable IDE0044 // Add readonly modifier

    /// <summary>
    /// Creates and manages Display objects also allowing the following optional functionality 
    /// Compositing - multiple camera render textures can be combined into a single texture
    /// Cubemap conversion - resulting textures can be processed to perform cubemap-equirectangular conversion
    /// Sharing - resulting textures can be shared via the TextureShareUtility
    /// </summary>
    public class CameraManager : MonoBehaviour
    {
        private List<IglooDisplay> displays = new List<IglooDisplay>();

        private const string LEFT_EYE = "Left";
        private const string RIGHT_EYE = "Right";

        internal IglooCameraData cameraData;
        private Dictionary<string, IglooCamera> cameras = new Dictionary<string, IglooCamera>();

        private RenderTexture stereoOutputTexture = null;

        public GameObject templateCamera;

        [System.Serializable]
        public class DisplayList
        {
            public IglooDisplay.IglooDisplayData[] Displays;
        }

        private void Start()
        {
            LoadSettings();
            // if not enabled in settings, do not do anything.
            if (!cameraData.isEnabled)
            {
                return;
            }
            if (cameraData.isStereo)
            {
                CreateCamera(LEFT_EYE, $"Igloo {cameraData.cameraTypeData.cameraType} Camera - Left Eye", cameraData.cameraTypeData.cameraType, -cameraData.stereoEyeDistance * 0.5f);
                CreateCamera(RIGHT_EYE, $"Igloo {cameraData} Camera - Right Eye", cameraData.cameraTypeData.cameraType, cameraData.stereoEyeDistance * 0.5f);
            } else
            {
                CreateCamera(LEFT_EYE, $"Igloo Camera - {cameraData.cameraTypeData.cameraType}", cameraData.cameraTypeData.cameraType, 0.0f);
            }

            AssignDisplayTextures();
            InitialiseTextureSenders();
            GetCommandLineSettings();
        }


        public void InitialiseTextureSenders ()
        {
            if (cameraData.isSpoutOutput)
            {
#if KLAK_SPOUT
                SpoutSender sender = gameObject.AddComponent<SpoutSender>();
                sender.SetResources(IglooManager.instance.spoutResources);
                sender.spoutName = cameraData.outputName;
                sender.captureMethod = Klak.Spout.CaptureMethod.Texture;
                Debug.Log($"<b>[Igloo]</b> Created Spout Sender with name {cameraData.outputName}");
                if (cameraData.isStereo)
                {
                    if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                    {
                        stereoOutputTexture = new RenderTexture(leftCamera.finalTexture.width, leftCamera.finalTexture.height * 2, 0)
                        {
                            name = "StereoOutputTexture",
                            wrapMode = TextureWrapMode.Repeat
                        };
                        sender.sourceTexture = stereoOutputTexture;
                    }
                }
                else
                {
                    if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                    {
                        sender.sourceTexture = leftCamera.finalTexture;
                    }
                }
#else
                Debug.LogWarning("<b>[Igloo]</b> Spout resources have not been applied, or the Klak Spout package has not been installed. Please select Igloo -> Apply Igloo Settings");
#endif
            }
            if (cameraData.isNDIOutput)
            {
#if KLAK_NDI
                NdiSender sender = gameObject.AddComponent<NdiSender>();
                sender.SetResources(IglooManager.instance.ndiResources);
                sender.ndiName = cameraData.outputName;
                sender.captureMethod = Klak.Ndi.CaptureMethod.Texture;
                Debug.Log($"<b>[Igloo]</b> Created NDI Sender with name {cameraData.outputName}");
                if (cameraData.isStereo)
                {
                    if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                    {
                        stereoOutputTexture = new RenderTexture(leftCamera.finalTexture.width, leftCamera.finalTexture.height * 2, 0)
                        {
                            name = "StereoOutputTexture",
                            wrapMode = TextureWrapMode.Repeat
                        };
                        sender.sourceTexture = stereoOutputTexture;
                    }
                }
                else
                {
                    if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                    {
                        sender.sourceTexture = leftCamera.finalTexture;
                    }
                }
#else
                Debug.LogWarning("<b>[Igloo]</b> NDI resources have not been applied, or the Klak NDI package has not been installed. Please select Igloo -> Apply Igloo Settings");
#endif
            }
        }

        /// <summary>
        /// Sets the Stereo state of the camera system, assumes the camera has been initialised in stereo mode. 
        /// </summary>
        /// <param name="state"></param>
        public void SetStereoState(bool state)
        {
            if (cameraData.isStereo)
            {
                if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                {
                    leftCamera.SetIPD(state ? -cameraData.stereoEyeDistance * 0.5f : 0);
                }
                if (cameras.TryGetValue(RIGHT_EYE, out var rightCamera))
                {
                    rightCamera.SetIPD(state ? cameraData.stereoEyeDistance * 0.5f : 0);

                    if (rightCamera)
                    {
                        if (rightCamera is CubemapCamera)
                        {
                            CubemapCamera rightCameraCube = (CubemapCamera)rightCamera;
                            foreach (var cam in rightCameraCube.cameras)
                            {
                                cam.enabled = state;
                            }
                        }
                        if (rightCamera is OffAxisCamera)
                        {
                            OffAxisCamera rightCameraOffAxis = (OffAxisCamera)rightCamera;
                            foreach (var cam in rightCameraOffAxis.cameras)
                            {
                                cam.camera.enabled = state;
                            }
                        }
                    }
                }
            } else
            {
                Debug.Log("IglooCameraManager - attempting to set the stereo state of a non-stereo camera will have no effect");
            }
        }

        public void SetIPD(float value)
        {
            if (cameraData.isStereo)
            {
                cameraData.stereoEyeDistance = value;
                if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                {
                    leftCamera.SetIPD(-value * 0.5f);
                }
                if (cameras.TryGetValue(RIGHT_EYE, out var rightCamera))
                {
                    leftCamera.SetIPD(value * 0.5f);
                }
            } else
            {
                Debug.Log("IglooCameraManager - attempting to set the IPD of a non-stereo camera will have no effect");
            }
        }

        void CreateCamera(string key, string name, IglooCameraType cameraType, float ipd = 0.0f)
        {
            GameObject camGO = new GameObject(name);
            camGO.transform.parent = this.transform;
            camGO.transform.localPosition = Vector3.zero;
            camGO.transform.localEulerAngles = Vector3.zero;
            switch (cameraType)
            {
                case IglooCameraType.CUBEMAP:
                    camGO.AddComponent<CubemapCamera>();
                    break;
                case IglooCameraType.OFFAXIS:
                    camGO.AddComponent<OffAxisCamera>();
                    break;
            }
            IglooCamera cam = camGO.GetComponent<IglooCamera>();
            cam.Setup(cameraData, templateCamera);
            cam.ipd = ipd;
            cameras[key] = cam;
        }

        void AssignDisplayTextures()
        {
            foreach (var display in displays)
            {
                if (cameras.TryGetValue(LEFT_EYE, out var leftCamera))
                {
                    display.leftEyeTexture = leftCamera.finalTexture;
                }

                if (cameras.TryGetValue(RIGHT_EYE, out var rightCamera))
                {
                    display.rightEyeTexture = rightCamera.finalTexture;
                }
                display.Setup();
            }
        }

        void LoadSettings()
        {
            cameraData = new IglooCameraData();

            string cameraSettingsPath = "";
            string localCameraSettingsPath = Application.streamingAssetsPath + "/IglooCameraSettings.json";
            string iceCameraSettingsPath = Utils.GetICEDataPath() + "/IglooCameraSettings.json";

            if (File.Exists(localCameraSettingsPath))
            {
                cameraSettingsPath = localCameraSettingsPath;
                Debug.Log($"<b>[Igloo]</b> Using Local Camera Settings found in Streaming Assets");

            } else if (File.Exists(iceCameraSettingsPath))
            {
                cameraSettingsPath = iceCameraSettingsPath;
                Debug.Log($"<b>[Igloo]</b> Using Global Camera Settings found in ICE Data");

            }

            if (!string.IsNullOrEmpty(cameraSettingsPath))
            {
                try
                {
                    string cameraJson = File.ReadAllText(cameraSettingsPath);
                    cameraData = JsonUtility.FromJson<IglooCameraData>(cameraJson);

                    if (cameraData.cameraTypeData.cameraType == IglooCameraType.OFFAXIS && cameraData.OffAxisData.cameras.Count == 0) {
                        List<TruePerspectiveXML.CameraConfig> truePespectiveCams = TruePerspectiveXML.Load(out cameraData.OffAxisData.unscaledCanvasWidth, out cameraData.OffAxisData.unscaledCanvasHeight);
                        if (truePespectiveCams == null)
                        {
                            Debug.LogWarning("<b>[Igloo]</b> TruePerspective configurations could not be loaded or the file is missing.");
                        }
                        if (truePespectiveCams.All(cam => cam is TruePerspectiveXML.PlaneConfig))
                        {
                            foreach (TruePerspectiveXML.PlaneConfig c in truePespectiveCams) {
                                IglooCameraData.OffAxisCameraData data = new IglooCameraData.OffAxisCameraData();
                                data.roi = new Vector4(c.ROI.x, (1f - c.ROI.y) - c.ROI.height, c.ROI.height, c.ROI.width);
                                data.lowerLeft  = new Vector3(c.PA.x, c.PA.y, -c.PA.z);
                                data.lowerRight = new Vector3(c.PB.x, c.PB.y, -c.PB.z);
                                data.upperLeft  = new Vector3(c.PC.x, c.PC.y, -c.PC.z);
                                cameraData.OffAxisData.cameras.Add(data);
                            }
                        }
                        else
                        {
                            Debug.LogWarning("<b>[Igloo]</b> There are non-PlaneConfig camera types in the file. All camera types must be Plane for Offaxis Cameras");
                        }
                    }

                } catch (Exception ex)
                {
                    Debug.LogWarning($"<b>[Igloo]</b> Error reading or parsing the file {cameraSettingsPath}. Exception: {ex.Message}");
                }
            } 
            else
            {
                Debug.Log($"<b>[Igloo]</b> Could not find any existing settings file using default instead");
                File.WriteAllText(Application.streamingAssetsPath + "/IglooCameraSettings.json", JsonUtility.ToJson(cameraData));
            }

            foreach (var entry in cameraData.CubemapData.faceList)
            {
                if (Enum.TryParse(entry.face, out Igloo.CubemapCamera.IglooCubemapFace faceEnumValue))
                {
                    cameraData.CubemapData.faces[faceEnumValue] = entry.enabled;
                }
            }

            DisplayList displayList = new DisplayList();
            string displaySettingsPath = "";
            string localDisplaySettingsPath = Application.streamingAssetsPath + "/IglooDisplaySettings.json";
            string iceDisplaySettingsPath = Utils.GetICEDataPath() + "/IglooDisplaySettings.json";

            if (File.Exists(localDisplaySettingsPath))
            {
                displaySettingsPath = localDisplaySettingsPath;
                Debug.Log($"<b>[Igloo]</b> Using Local Display Settings found in Streaming Assets");
            }
            else if (File.Exists(iceDisplaySettingsPath))
            {
                displaySettingsPath = iceDisplaySettingsPath;
                Debug.Log($"<b>[Igloo]</b> Using Global Display Settings found in ICE Data");
            }

            if (!string.IsNullOrEmpty(displaySettingsPath))
            {
                try
                {
                    string displayJson = File.ReadAllText(displaySettingsPath);
                    displayList = JsonUtility.FromJson<DisplayList>(displayJson);
                    if(displayList.Displays.Length != 0) CreateDisplays(displayList.Displays);
                } 
                catch (Exception ex)
                {
                    Debug.LogWarning($"<b>[Igloo]</b> Error reading or parsing the display file {displaySettingsPath}. Exception: {ex.Message}");
                }
            } 
            else
            {
                Debug.Log($"<b>[Igloo]</b> Could not find any existing settings file using default instead");
                File.WriteAllText(Application.streamingAssetsPath + "/IglooDisplaySettings.json", JsonUtility.ToJson(displayList));
            }
        }

        void CreateDisplays(IglooDisplay.IglooDisplayData[] displayDataArray)
        {
            foreach (var data in displayDataArray)
            {
                GameObject displayObject = new GameObject($"Display_{data.targetDisplay}");
                displayObject.transform.parent = this.transform;
                IglooDisplay iglooDisplay = displayObject.AddComponent<IglooDisplay>();
                iglooDisplay.SetNormalizedSourceX(data.SourceX);
                iglooDisplay.SetNormalizedSourceY(data.SourceY);
                iglooDisplay.SetNormalizedSourceWidth(data.SourceWidth);
                iglooDisplay.SetNormalizedSourceHeight(data.SourceHeight);
                iglooDisplay.SetTargetDisplay(data.targetDisplay);
                iglooDisplay.SetStereoMode((IglooDisplay.STEREO_MODE)data.StereoMode);
                displays.Add(iglooDisplay);
            }

            for (int i = 1; i < Display.displays.Length; i++)
            {
                Display.displays[i].Activate();
            }
        }

        void Update()
        {
            if (cameraData.isStereo && cameraData.isEnabled)
            {
                if (IsReadyForStereoOutput())
                {
                    CopyTexturesToStereoOutput();
                }
            }
        }

        bool IsReadyForStereoOutput()
        {
            return stereoOutputTexture != null
                && cameras[LEFT_EYE].finalTexture != null
                && cameras[RIGHT_EYE].finalTexture != null;
        }


        void CopyTexturesToStereoOutput()
        {
            int width = cameras[RIGHT_EYE].finalTexture.width;
            int height = cameras[RIGHT_EYE].finalTexture.height;

            Graphics.CopyTexture(cameras[RIGHT_EYE].finalTexture, 0, 0, 0, 0, width, height, stereoOutputTexture, 0, 0, 0, 0);
            Graphics.CopyTexture(cameras[LEFT_EYE].finalTexture, 0, 0, 0, 0, width, height, stereoOutputTexture, 0, 0, 0, height);
        }

        void GetCommandLineSettings()
        {
            string stereoState = string.Empty;
            var args = System.Environment.GetCommandLineArgs();
            for (int i = 0; i < args.Length; i++)
            {
                if (args[i].Equals("-setStereoState", System.StringComparison.InvariantCultureIgnoreCase)) stereoState = args[i + 1];
            }

            if (bool.TryParse(stereoState, out bool state))
            {
                SetStereoState(state);
            }
        }

        internal void SetDisplaysEnabled(bool val)
        {
            foreach(IglooDisplay display in displays)
            {
                display.enabled = val;
            }
        }

        public RenderTexture GetOutputTexture()
        {
            if (cameras.ContainsKey(LEFT_EYE)) return cameras[LEFT_EYE].finalTexture;
            else return null;
        }
    }
}
