using Igloo;
using Igloo.Common;
using System;
using UnityEngine;
#if IGLOO_HDRP
using UnityEngine.Rendering.HighDefinition;
#endif

namespace Igloo {

    public class CubemapCamera : IglooCamera
    {
        [System.Serializable]
        public enum IglooCubemapFace { Left, Front, Right, Back, Down, Up }

        public bool useTruePerspective = false;
        public int  cubemapFaceResolution = 2000;
        private bool isSetup = false;

        public Camera[] cameras = new Camera[6];
        private RenderTexture[] renderTextures = new RenderTexture[6];

        public float cubeSize = 1.0f;

        public GameObject head;

        private Vector3 GetLowerLeft(int index)  { return (lowerLeftArray[index]  * cubeSize) + transform.position; }
        private Vector3 GetLowerRight(int index) { return (lowerRightArray[index] * cubeSize) + transform.position; }
        private Vector3 GetUpperLeft(int index)  { return (upperLeftArray[index]  * cubeSize) + transform.position; }

        private Vector3 GetUpperRight(int index)
        {
            return GetUpperLeft(index) + (GetLowerRight(index) - GetLowerLeft(index));
        }

        // Lower Left
        private Vector3[] lowerLeftArray = new Vector3[6] {
        new Vector3(-1, -1, -1),
        new Vector3(-1, -1, 1),
        new Vector3(1, -1, 1),
        new Vector3(1, -1, -1),
        new Vector3(-1, -1, -1),
        new Vector3(-1, -1, -1),
        };
        // Lower Right
        private Vector3[] lowerRightArray = new Vector3[6] {
        new Vector3(-1, -1, 1),
        new Vector3(1, -1, 1),
        new Vector3(1, -1, -1),
        new Vector3(-1, -1, -1),
        new Vector3(1, -1, -1),
        new Vector3(1, -1, -1),
        };
        // Upper Left
        private Vector3[] upperLeftArray = new Vector3[6] {
        new Vector3(-1, 1, -1),
        new Vector3(-1, 1, 1),
        new Vector3(1, 1, 1),
        new Vector3(1, 1, -1),
        new Vector3(-1,-1, 1),
        new Vector3(-1, -1, 1),
        };

        public override void Setup(IglooCameraData data, GameObject templateCameraGameObject = null)
        {
            base.Setup(data);

            useTruePerspective    = data.CubemapData.useTruePerspective;
            useWarpAndBlend       = data.useWarpAndBlend;
            cubemapFaceResolution = data.CubemapData.cubemapFaceResolution;

            effectsChain = gameObject.AddComponent<EffectsChain>();

            CubemapToEquirectangularEffect cubemapToEquirectangularEffect = new CubemapToEquirectangularEffect();
            cubemapToEquirectangularEffect.SetOutputResolution(new Vector2Int(cubemapFaceResolution * 4, cubemapFaceResolution * 2));
            effectsChain.AddEffect(cubemapToEquirectangularEffect);


            if (useTruePerspective && Utils.ShadersLoaded)
            {
                TruePerspectiveEffect truePerspectiveEffect = new TruePerspectiveEffect();
                effectsChain.AddEffect(truePerspectiveEffect);
            }

            if (useWarpAndBlend && Utils.ShadersLoaded)
            {
                WarpBlendEffect warpBlendEffect = new WarpBlendEffect();
                effectsChain.AddEffect(warpBlendEffect);
            }
            
            foreach (var face in data.CubemapData.faces)
            {
                if (face.Value)
                {
                    GameObject cameraGO;
                    if (templateCameraGameObject != null)
                    {
                        cameraGO = Instantiate(templateCameraGameObject, this.transform);
                        cameraGO.name = "Face_" + face.Key.ToString();
                        if (!cameraGO.GetComponent<Camera>()) cameraGO.AddComponent<Camera>();
                        if (cameraGO.GetComponent<AudioListener>()) Destroy(cameraGO.GetComponent<AudioListener>());
                        cameraGO.SetActive(true);
                    }
                    else
                    {
                        cameraGO = new GameObject("Face_" + face.Key.ToString());
                        cameraGO.AddComponent<Camera>(); 
                    }
                   
                    Camera camera = cameraGO.GetComponent<Camera>();
                    camera.fieldOfView             = 90;
                    camera.nearClipPlane           = data.nearClipPlane;
                    camera.farClipPlane            = data.farClipPlane;
                    camera.transform.parent        = this.gameObject.transform;
                    camera.transform.localPosition = Vector3.zero;
                    camera.transform.rotation      = Quaternion.identity;
                    camera.enabled                 = true;

#if IGLOO_HDRP
                    if (HDDynamicResolutionPlatformCapabilities.DLSSDetected)
                    {
                        if (camera.gameObject.AddComponent<HDAdditionalCameraData>())
                        {
                            camera.allowDynamicResolution = true;
                            HDAdditionalCameraData HDC = camera.GetComponent<HDAdditionalCameraData>();   
                            HDC.allowDynamicResolution = true;
                            HDC.allowDeepLearningSuperSampling = true;
                            HDC.deepLearningSuperSamplingUseOptimalSettings = true;
                        }
                    }
#endif

                    RenderTexture renderTexture = new RenderTexture(cubemapFaceResolution, cubemapFaceResolution, 24);
                    renderTexture.dimension = UnityEngine.Rendering.TextureDimension.Tex2D;
                    renderTexture.hideFlags = HideFlags.HideAndDontSave;
                    camera.targetTexture = renderTexture;

                    cameras[(int)face.Key] = camera;
                    renderTextures[(int)face.Key] = renderTexture;
                }
            }

            if (renderTextures[1] != null) cubemapToEquirectangularEffect.SetCubemapFace((IglooCubemapFace.Left), renderTextures[1]);
            if (renderTextures[2] != null) cubemapToEquirectangularEffect.SetCubemapFace((IglooCubemapFace.Front), renderTextures[2]);
            if (renderTextures[3] != null) cubemapToEquirectangularEffect.SetCubemapFace((IglooCubemapFace.Right), renderTextures[3]);
            if (renderTextures[0] != null) cubemapToEquirectangularEffect.SetCubemapFace((IglooCubemapFace.Back), renderTextures[0]);
            if (renderTextures[5] != null) cubemapToEquirectangularEffect.SetCubemapFace((IglooCubemapFace.Down), renderTextures[5]);
            if (renderTextures[4] != null) cubemapToEquirectangularEffect.SetCubemapFace((IglooCubemapFace.Up), renderTextures[4]);
            


            if (cameras[0] != null) cameras[0].transform.localRotation = Quaternion.Euler(0f, -90f, 0f); // Left
            if (cameras[1] != null) cameras[1].transform.localRotation = Quaternion.Euler(0f, 0f, 0f);   // Front
            if (cameras[2] != null) cameras[2].transform.localRotation = Quaternion.Euler(0f, 90f, 0f);  // Right
            if (cameras[3] != null) cameras[3].transform.localRotation = Quaternion.Euler(0f, 180f, 0f); // Back
            if (cameras[4] != null) cameras[4].transform.localRotation = Quaternion.Euler(90f, 0f, 0f);  // Down
            if (cameras[5] != null) cameras[5].transform.localRotation = Quaternion.Euler(-90f, 0f, 0f); // Up


            finalTexture = effectsChain.ApplyEffects(finalTexture);

            this.transform.localEulerAngles = data.cameraRotationOffset;

            head = IglooManager.instance.HeadManager.gameObject;
            isSetup = true;
        }

        public override void SetIPD(float val)
        {
            ipd = val;
            if (cameras[0] != null)cameras[0].transform.localPosition  = new Vector3(0f, 0f, -ipd);  // Left 
            if (cameras[1] != null)cameras[1].transform.localPosition  = new Vector3(-ipd, 0f, 0f);  // Front
            if (cameras[2] != null)cameras[2].transform.localPosition  = new Vector3(0f, 0f, ipd);   // Right
            if (cameras[3] != null)cameras[3].transform.localPosition  = new Vector3(ipd, 0f, 0f);   // Back
            if (cameras[4] != null)cameras[4].transform.localPosition  = new Vector3(-ipd, 0, 0f);   // Down
            if (cameras[5] != null)cameras[5].transform.localPosition  = new Vector3(-ipd, 0f, 0f);  // Up

        }

        private void FixedUpdate()
        {
            if (head) transform.localPosition = head.transform.localPosition;

        }
        private void  LateUpdate()
        { 
            if (!isSetup) return;

            SetIPD(ipd);

            for (int i = 0; i < cameras.Length; i++)
            {
                if (cameras[i] != null) cameras[i].projectionMatrix = Utils.GetAsymProjMatrix(GetLowerLeft(i), GetLowerRight(i), GetUpperLeft(i), cameras[i].gameObject.transform.position, ncp, fcp);
            }

            // Input is not required here, maybe better to get a tempoary texture
            finalTexture = effectsChain.ApplyEffects(finalTexture);


            foreach (var effect in effectsChain.effects)
            {
                if (effect is TruePerspectiveEffect)
                {
                    TruePerspectiveEffect tp = (TruePerspectiveEffect)effect;
                    tp.SetEyePosition(this.transform.localPosition);
                }
            }
        }

        void OnDrawGizmos()
        {
            for (int i = 0; i < cameras.Length; i++)
            {
                Gizmos.color = Color.blue;

                Vector3 lowerLeft = GetLowerLeft(i);
                Vector3 lowerRight = GetLowerRight(i);
                Vector3 upperLeft = GetUpperLeft(i);
                Vector3 upperRight = GetUpperRight(i);

                Gizmos.DrawLine(upperLeft, lowerLeft);
                Gizmos.DrawLine(lowerLeft, lowerRight);
                Gizmos.DrawLine(lowerRight, upperRight);
                Gizmos.DrawLine(upperRight, upperLeft);

                Gizmos.color = Color.yellow;
                float radius = Vector3.Distance(lowerRight, lowerLeft)/2f;
                Gizmos.DrawWireSphere(this.transform.position, radius);
            }
        }

    }
}



