using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Xml.Linq;
using System.Linq;
using UnityEngine;

namespace Igloo
{
    public abstract class Effect
    {
        protected Material effectMaterial_;
        protected RenderTexture output_;

        public abstract void Setup();

        public abstract RenderTexture ApplyEffect(RenderTexture source);
    }

    public class CubemapToEquirectangularEffect : Effect
    {      
        private float horizontalFOV_    = 360f;
        private float verticalFOV_      = 180f;
        private Vector2Int outputResolution_ = new Vector2Int(8000, 4000);
        public void SetOutputResolution(Vector2Int resolution)
        {
            outputResolution_ = resolution;
        }
        private Dictionary<Igloo.CubemapCamera.IglooCubemapFace, RenderTexture> cubemapFaces = new Dictionary<Igloo.CubemapCamera.IglooCubemapFace, RenderTexture>();

        public override void Setup()
        {
            effectMaterial_ = new Material(Shader.Find("Hidden/Igloo/Equirectangular"));
            effectMaterial_.SetFloat("_HorizontalFov", horizontalFOV_ * Mathf.Deg2Rad);
            effectMaterial_.SetFloat("_VerticalFov", verticalFOV_ * Mathf.Deg2Rad);
            output_ = new RenderTexture(outputResolution_.x, outputResolution_.y, 24);
            output_.wrapMode = TextureWrapMode.Repeat;
        }

        public override RenderTexture ApplyEffect(RenderTexture source)
        {
            Graphics.Blit(null, output_, effectMaterial_);
            return output_;
        }

        public void SetCubemapFace(Igloo.CubemapCamera.IglooCubemapFace face, RenderTexture newTexture)
        {
            cubemapFaces[face] = newTexture;
            effectMaterial_.SetTexture(GetFaceName((int)face), newTexture);
        }

        private static readonly string[] FaceNames =
        {
            "_FaceTexPZ",
            "_FaceTexPX",
            "_FaceTexNZ",
            "_FaceTexNX",
            "_FaceTexNY",
            "_FaceTexPY"
        };

        private string GetFaceName(int i)
        {
            if (i >= 0 && i < FaceNames.Length)
                return FaceNames[i];

            return "null";
        }
    }

    public class TruePerspectiveLookupEffect : Effect
    {
        private Texture2D warpTex;

        public override void Setup()
        {
            string programDataPath = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
            string warpPath = Path.Combine(programDataPath, "Igloo Vision\\IglooCoreEngine\\settings\\TruePerspectiveMap32.bmp");
            int tpWidth = 0;
            int tpHeight = 0;

            using (FileStream fs = new FileStream(warpPath, FileMode.Open, FileAccess.Read))
            {
                byte[] header = new byte[54];
                fs.Read(header, 0, 54);
                tpWidth = BitConverter.ToInt32(header, 18);
                tpHeight = BitConverter.ToInt32(header, 22);
            }

            warpTex = Utils.LoadFloatTexture(warpPath, tpWidth, tpHeight, 54);

            if (warpTex != null)
            {
                effectMaterial_ = new Material(Utils.IglooShaders["Igloo/TruePerspective"]);
                effectMaterial_.SetTexture("_WarpTex", warpTex);
                output_ = new RenderTexture(tpWidth, tpHeight, 0);
            }
            else
            {
                UnityEngine.Debug.LogError("<b>[Igloo]</b> Failed to load Warp texture from " + warpPath);
            }
        }

        public override RenderTexture ApplyEffect(RenderTexture source)
        {
            Graphics.Blit(source, output_, effectMaterial_);
            return output_;
        }

    }

    public class TruePerspectiveEffect : Effect
    {
        public int outputWidth = 1000;
        public int outputHeight = 1000;
        private bool isSetup = false;

        public List<TruePerspectiveXML.CameraConfig> cameraConfigs = null;

        public override void Setup()
        {
            cameraConfigs = TruePerspectiveXML.Load(out outputWidth, out outputHeight);
            if (cameraConfigs == null)
            {
                Debug.LogWarning("<b>[Igloo]</b> Camera configurations could not be loaded or the configuration file is missing.");
                return;
            }
            if (cameraConfigs.OfType<TruePerspectiveXML.PlaneConfig>().Any()) 
            {
                Debug.LogWarning("<b>[Igloo]</b> Found one or more PlaneConfig configurations in the loaded camera configurations. Use Offaxis camera instead");
                return;
            }
      
            foreach (TruePerspectiveXML.CameraConfig config in cameraConfigs)
            {
                float w = outputWidth * config.ROI.width;
                float h = outputHeight * config.ROI.height;

                config.output_ = new RenderTexture ((int)w, (int)h, 0);
                config.output_.wrapMode = TextureWrapMode.Repeat;
                config.LoadMaterial();
                
                if (config is TruePerspectiveXML.CylinderConfig) {
                    TruePerspectiveXML.CylinderConfig cc = (TruePerspectiveXML.CylinderConfig)config;
                    cc.material_.SetFloat("_u_cylinder_height", cc.Height);
                    cc.material_.SetFloat("_u_cylinder_radius", cc.Radius);
                    cc.material_.SetFloat("_u_cylinder_theta_range", cc.ThetaRange * ((2.0f * Mathf.PI) / 360.0f));
                    cc.material_.SetFloat("_u_cylinder_theta_offset", cc.ThetaOffset * ((2.0f * Mathf.PI) / 360.0f));
                  
                }
                else if (config is TruePerspectiveXML.DiskConfig)
                {
                    TruePerspectiveXML.DiskConfig dc = (TruePerspectiveXML.DiskConfig)config;
                    dc.material_.SetVector("_u_disk_pa", new Vector3(dc.PA.x, dc.PA.y, -dc.PA.z));
                    dc.material_.SetVector("_u_disk_pb", new Vector3(dc.PB.x, dc.PB.y, -dc.PB.z));
                    dc.material_.SetVector("_u_disk_pc", new Vector3(dc.PC.x, dc.PC.y, -dc.PC.z));
                    dc.material_.SetFloat("_u_disk_radius", dc.Radius);
                    dc.material_.SetFloat("_u_disk_theta_offset", dc.ThetaOffset * ((2.0f * Mathf.PI) / 360.0f));
                }
            }

            output_ = new RenderTexture(outputWidth, outputHeight, 0);
            output_.wrapMode = TextureWrapMode.Repeat;

            isSetup = true;
        }

        public override RenderTexture ApplyEffect(RenderTexture source)
        {
            if (!isSetup)
            {
                return CreateColoredTexture(100, 100, Color.red);

            }
            if (!Utils.ShadersLoaded)
            {
                return CreateColoredTexture(100, 100, new Color(1f, 0f, 1f));
            }
            foreach (TruePerspectiveXML.CameraConfig config in cameraConfigs)
            {
                Graphics.Blit(source, config.output_, config.material_);
                int h = (int)(config.ROI.height * output_.height);
                int y = (int)(((1.0f - config.ROI.y) - config.ROI.height) * outputHeight);
                if (config is TruePerspectiveXML.DiskConfig)
                {
                    TruePerspectiveXML.DiskConfig diskConfig = (TruePerspectiveXML.DiskConfig) config;
                    Graphics.Blit(diskConfig.output_, diskConfig.scaledOutputTexture, diskConfig.scaleMaterial);
                    Graphics.CopyTexture(diskConfig.scaledOutputTexture, 0, 0, 0, 0, diskConfig.scaledOutputTexture.width, diskConfig.scaledOutputTexture.height, output_, 0, 0, (int)(config.ROI.x * output_.width), y);
                }
                else Graphics.CopyTexture(config.output_, 0, 0, 0, 0, config.output_.width, config.output_.height, output_, 0, 0, (int)(config.ROI.x * output_.width), y);

            }
            return output_;
        }

        public void SetEyePosition(Vector3 eyePos) {
            if (!isSetup) return;
            foreach (TruePerspectiveXML.CameraConfig config in cameraConfigs)
            {
                config.material_.SetVector("_u_eye", eyePos);
            }
        }

        private RenderTexture CreateColoredTexture(int width, int height, Color color)
        {
            RenderTexture coloredTexture = new RenderTexture(width, height, 0);
            coloredTexture.wrapMode = TextureWrapMode.Repeat;
            Material fillMaterial = new Material(Shader.Find("Hidden/Internal-Colored"));
            fillMaterial.SetColor("_Color", color);
            Graphics.Blit(null, coloredTexture, fillMaterial);
            return coloredTexture;
        }
    }

    public class WarpBlendEffect : Effect
    {
        private Texture2D warpTex;
        private Texture2D blendTex;
        private Texture2D canvasMaskTex;


        public override void Setup()
        {
            string programDataPath = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
            string warpPath        = Path.Combine(programDataPath, "Igloo Vision\\IglooCoreEngine\\warping\\Screen-warp32.bmp");
            string blendPath       = Path.Combine(programDataPath, "Igloo Vision\\IglooCoreEngine\\warping\\Screen-edgeBlend.png");
            string canvasMaskPath  = Path.Combine(programDataPath, "Igloo Vision\\IglooCoreEngine\\warping\\canvasMask.png");

            int tpWidth = 0;
            int tpHeight = 0;

            using (FileStream fs = new FileStream(warpPath, FileMode.Open, FileAccess.Read))
            {
                byte[] header = new byte[54];
                fs.Read(header, 0, 54);
                tpWidth = BitConverter.ToInt32(header, 18);
                tpHeight = BitConverter.ToInt32(header, 22);
            }

            warpTex = Utils.LoadFloatTexture(warpPath, tpWidth, tpHeight, 54);

            if (warpTex != null)
            {
                effectMaterial_ = new Material(Shader.Find("Igloo/Warp"));
                effectMaterial_.SetTexture("_WarpTex", warpTex);
                output_ = new RenderTexture(tpWidth, tpHeight, 0);
            }
            else
            {
                UnityEngine.Debug.LogError("<b>[Igloo]</b> Failed to load Warp texture from " + warpPath);
            }

            blendTex = LoadTexture(blendPath);
            if (blendTex != null)
            {
                effectMaterial_.SetTexture("_BlendTex", blendTex);
            }

            canvasMaskTex = LoadTexture(canvasMaskPath);
            if (canvasMaskTex != null)
            {
                effectMaterial_.SetTexture("_CanvasMaskTex", canvasMaskTex);
            }

            effectMaterial_.SetFloat("_canvasPosU", 0f);
            effectMaterial_.SetFloat("_canvasWidth", 1f);
            effectMaterial_.SetFloat("_overlapLeft", 0f);
            effectMaterial_.SetFloat("_overlapRight", 0f);
            effectMaterial_.SetFloat("_stereoMode", 0f);
            effectMaterial_.SetFloat("_displayPosX", 0f);
            effectMaterial_.SetFloat("_displayPosY", 0f);
            effectMaterial_.SetFloat("_displayScaleX", 1f);
            effectMaterial_.SetFloat("_displayScaleY", 1f);
        }

        public override RenderTexture ApplyEffect(RenderTexture source)
        {
            Graphics.Blit(source, output_, effectMaterial_);
            return output_;
        }

        Texture2D LoadTexture(string path)
        {
            byte[] fileData;
            Texture2D tex = null;
            if (File.Exists(path))
            {
                _ = new Texture2D(2, 2);
                fileData = File.ReadAllBytes(path);
                tex = new Texture2D(2, 2);
                tex.LoadImage(fileData);
            }
            return tex;
        }

    }

    public class EffectsChain : MonoBehaviour
    {
        public RenderTexture outputTexture;
        public List<Effect> effects;

        EffectsChain ()
        {
            effects = new List<Effect>();
        }

        public void AddEffect(Effect effect)
        {
            effects.Add(effect);
            effect.Setup();
        }

        public void RemoveEffect(Effect effect)
        {
            effects.Remove(effect);
        }

        public RenderTexture ApplyEffects(RenderTexture source)
        {
            foreach (Effect effect in effects)
            {
                source = effect.ApplyEffect(source);
            }

            if (outputTexture == null && source != null) CreateOrResizeOutputTexture(source.width, source.height);
            if (outputTexture != null && source != null)
            {
                if (outputTexture.width != source.width || outputTexture.height != source.height)
                {
                    CreateOrResizeOutputTexture(source.width, source.height);
                }

                Graphics.Blit(source, outputTexture);
            }

            return outputTexture;
        }

        private void CreateOrResizeOutputTexture(int width, int height)
        {
            if (outputTexture != null)
            {
                if (outputTexture.width == width && outputTexture.height == height)
                {
                    return;
                }
                Debug.Log("<b>[Igloo]</b> Output texture size mismatch. Destroying and recreating texture.");
                Debug.LogFormat("<b>[Igloo]</b> Old Size: {0}x{1}, New Size: {2}x{3}", outputTexture.width, outputTexture.height, width, height);
                Destroy(outputTexture);
            }
            else
            {
                Debug.Log("<b>[Igloo]</b> Output texture does not exist. Creating a new texture.");
            }

            outputTexture = new RenderTexture(width, height, 24);
            outputTexture.wrapMode = TextureWrapMode.Repeat;
            outputTexture.name = $"{gameObject.name} - Effects Chain Output";
            Debug.LogFormat("<b>[Igloo]</b> Output texture created with resolution: {0}x{1}.", outputTexture.width, outputTexture.height);
        }

    }
}
