﻿using UnityEngine;
using System.Net.Sockets;
using System.Net;

namespace Igloo.Common
{
#pragma warning disable IDE0090 // Use New()...
    /// <summary>
    /// The Igloo Head Manager class
    /// </summary>
    public class HeadManager : Singleton<HeadManager>
    {
        /// <summary>
        /// Head Tracking System Type
        /// 0 - Default
        /// 1 - Optitrack 
        /// 2 - VRPN
        /// </summary>
        public enum HeadTrackingInput { Default, Optitrack, VRPN };

        /// <summary>
        /// Current Head Tracking Input System
        /// </summary>
        public HeadTrackingInput headTrackingInput = HeadTrackingInput.Default;

        /// <summary>
        /// If True, Head Tracking is enabled
        /// </summary>
        private bool isHeadTracking = false;

        /// <summary>
        /// Positon Cache for last update
        /// </summary>
        internal Vector3 headPositionOffset = Vector3.zero;

        /// <summary>
        /// Rotation Cache for last update
        /// </summary>
        internal Vector3 headRotationOffset = Vector3.zero;

        /// <summary>
        /// Called when any settings have been changed for the HeadManager
        /// </summary>
        public HeadSettingsChange OnHeadSettingsChange;

        /// <summary>
        /// The server address of the Optitrack machine, default 127.0.0.1
        /// </summary>
        public string optitrackServerIPAddress = "127.0.0.1";

        /// <summary>
        /// The rigidbody ID that the optitrack system is assigned to follow
        /// </summary>
        public int optitrackHeadRigidBodyID = 1;

        /// <summary>
        /// Delegate for HeadSettingsChange
        /// </summary>
        public delegate void HeadSettingsChange();

        /// <summary>
        /// Applied offset to incoming tracking data
        /// </summary>
        public Vector3 trackingPositionOffset = Vector3.zero;

        /// <summary>
        /// Creates a camera as a child of the Head Manager
        /// </summary>
        /// <returns></returns>
        public Camera CreateCamera()
        {
            GameObject cam = new GameObject
            {
                name = "IglooMainCamera"
            };
            cam.AddComponent <Camera>();
            cam.transform.parent = transform;
            cam.transform.localEulerAngles = Vector3.zero;
            cam.transform.localPosition = Vector3.zero;
            
            if(!FindObjectOfType<AudioListener>()) cam.AddComponent<AudioListener>();
#if IGLOO_PLAYER
            cam.AddComponent<PlayerPointer>();
#endif
            return cam.GetComponent<Camera>(); 
        }

        /// <summary>
        /// Returns HeadTracking Enabled whilst calling OnHeadSettingsChange
        /// </summary>
        public bool HeadTracking
        {
            get { return isHeadTracking; }
            set
            {
                isHeadTracking = value;
                if (!isHeadTracking && !IglooManager.instance.followObject)
                {
                    transform.localPosition = headPositionOffset;
                    transform.localEulerAngles = headRotationOffset;
                }
            }
        }

        /// <summary>
        /// Near Clip Plane for created cameras
        /// </summary>
        public float NearClipPlane { get { return nearClipPlane; } set { nearClipPlane = value; } }
        private float nearClipPlane = 0.03f;

        /// <summary>
        /// Far clip plane for created cameras
        /// </summary>
        public float FarClipPlane { get { return farClipPlane; } set { farClipPlane = value; } }
        private float farClipPlane = 1000f;

        /// <summary>
        /// Writes the current Head Settings to the Igloo Settings xml
        /// </summary>
        /// <param name="hs">A populated Head Settings class</param>
        public void SetSettings(HeadSettings hs)
        {
            transform.localPosition = hs.headPositionOffset.Vector3;
            headPositionOffset = hs.headPositionOffset.Vector3;
            transform.localEulerAngles = hs.headRotationOffset.Vector3;
            headRotationOffset = hs.headRotationOffset.Vector3;
            isHeadTracking = hs.headtracking;
            if (hs.optitrackServerIP != "") optitrackServerIPAddress = hs.optitrackServerIP;
            if (hs.optitrackHeadRigidBodyID != 0) optitrackHeadRigidBodyID = hs.optitrackHeadRigidBodyID;
            headTrackingInput = (HeadTrackingInput)hs.headTrackingInput;

            NetworkManagerOSC.instance.OnHeadPosition += HandlePositonMessage;
            NetworkManagerOSC.instance.OnHeadRotation += HandleRotationMessage;
            if (isHeadTracking) 
            { 
                if (headTrackingInput == HeadTrackingInput.Optitrack)
                {
                    Igloo.Controllers.OptitrackRigidBodyIgloo optitrack = this.gameObject.AddComponent<Igloo.Controllers.OptitrackRigidBodyIgloo>();
                    optitrack.Setup(optitrackHeadRigidBodyID, true, false, optitrackServerIPAddress, GetLocalIPAddress());
                    optitrack.trackingPositionOffset = hs.trackingPositionOffset.Vector3;
                } else if (headTrackingInput == HeadTrackingInput.VRPN)
                {
                    Igloo.VRPN.TrackerSettings vrpn = this.gameObject.AddComponent<Igloo.VRPN.TrackerSettings>();
                    vrpn.Setup("Head", 0, true, false, optitrackServerIPAddress);
                }
            }
        }

        /// <summary>
        /// Handles an OSC message with a new position vector
        /// </summary>
        /// <param name="pos">Vector3, New Position</param>
        public virtual void HandlePositonMessage(Vector3 pos)
        {
            if (isHeadTracking && headTrackingInput != HeadTrackingInput.Optitrack) transform.localPosition = pos;
        }

        public string GetLocalIPAddress() {
            var host = Dns.GetHostEntry(Dns.GetHostName());
            foreach (var ip in host.AddressList) {
                if (ip.AddressFamily == AddressFamily.InterNetwork) {
                    return ip.ToString();
                }
            }
            throw new System.Exception("No network adapters with an IPv4 address in the system!");
        }

        /// <summary>
        /// Handles an OSC message with a new rotation vector
        /// </summary>
        /// <param name="rotation">Vector3, New Euler Rotation</param>
        public virtual void HandleRotationMessage(Vector3 rotation)
        {
            if (isHeadTracking && headTrackingInput != HeadTrackingInput.Optitrack) transform.localEulerAngles = rotation;
        }

        /// <summary>
        /// Reads the Head Settings of the Igloo Settings XML
        /// </summary>
        /// <returns>A populated Head Settings class</returns>
        public HeadSettings GetSettings()
        {
            HeadSettings hs = new HeadSettings {
                // Use starting head position as head tracking is not friendly here
                headPositionOffset = new Vector3Item(headPositionOffset),
                headRotationOffset = new Vector3Item(headRotationOffset),
                optitrackServerIP = optitrackServerIPAddress,
                optitrackHeadRigidBodyID = optitrackHeadRigidBodyID,
                headtracking = isHeadTracking,
                headTrackingInput = (int)headTrackingInput,
                trackingPositionOffset = new Vector3Item(trackingPositionOffset),
            };
            return hs;
        }
    }
}

