using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace Igloo
{
    [RequireComponent(typeof(Camera))]
    public class IglooDisplay : MonoBehaviour
    {
        [System.Serializable]
        public class IglooDisplayData
        {
            public int targetDisplay;
            public float SourceX;
            public float SourceY;
            public float SourceWidth;
            public float SourceHeight;
            public int StereoMode = 0;
        }

        public Material stereoSBSMaterial;
        public Material monoMaterial;

        [Range(0, 1)]
        public float sourceX = 0;
        [Range(0, 1)]
        public float sourceY = 0;
        [Range(0, 1)]
        public float sourceWidth = 1;
        [Range(0, 1)]
        public float sourceHeight = 1;

        // 0 - Off, 1- SBS
        public enum STEREO_MODE {MONO,SBS}
        public STEREO_MODE stereoMode = STEREO_MODE.MONO;

        public RenderTexture leftEyeTexture;
        public RenderTexture rightEyeTexture;

        public void Setup()
        {
            switch (stereoMode) {
                case STEREO_MODE.MONO:
                    monoMaterial = new Material(Shader.Find("Igloo/SubsectionRender"));
                    monoMaterial.SetTexture("_MainTex", leftEyeTexture);
                    break;
                case STEREO_MODE.SBS:
                    stereoSBSMaterial = new Material(Shader.Find("Igloo/SideBySide"));
                    stereoSBSMaterial.SetTexture("_LeftEyeTexture", leftEyeTexture);
                    stereoSBSMaterial.SetTexture("_RightEyeTexture", rightEyeTexture);
                    break;
            }
        }

        private void OnRenderImage(RenderTexture source, RenderTexture destination)
        {
            switch (stereoMode)
            {
                case STEREO_MODE.MONO:
                    if (monoMaterial != null)
                    {
                        monoMaterial.SetTexture("_MainTex"   , leftEyeTexture);
                        monoMaterial.SetFloat("_SourceX"     , sourceX);
                        monoMaterial.SetFloat("_SourceY"     , sourceY);
                        monoMaterial.SetFloat("_SourceWidth" , sourceWidth);
                        monoMaterial.SetFloat("_SourceHeight", sourceHeight);
                        Graphics.Blit(null, destination, monoMaterial);
                    }
                    break;
                case STEREO_MODE.SBS:
                    if (stereoSBSMaterial != null)
                    {
                        stereoSBSMaterial.SetTexture("_LeftEyeTexture" , leftEyeTexture);
                        stereoSBSMaterial.SetTexture("_RightEyeTexture", rightEyeTexture);
                        stereoSBSMaterial.SetFloat("_SourceX"          , sourceX);
                        stereoSBSMaterial.SetFloat("_SourceY"          , sourceY);
                        stereoSBSMaterial.SetFloat("_SourceWidth"      , sourceWidth);
                        stereoSBSMaterial.SetFloat("_SourceHeight"     , sourceHeight);
                        Graphics.Blit(null, destination, stereoSBSMaterial);
                    }
                    break;
            }
        }

        private void OnValidate()
        {
            UpdateShaderProperties();
        }

        public void SetNormalizedSourceX(float value)
        {
            sourceX = value;
        }

        public void SetNormalizedSourceY(float value)
        {
            sourceY = value;
        }

        public void SetNormalizedSourceWidth(float value)
        {
            sourceWidth = value;
        }

        public void SetNormalizedSourceHeight(float value)
        {
            sourceHeight = value;
        }

        public void SetTargetDisplay(int value)
        {
            GetComponent<Camera>().targetDisplay = value;
        }

        public void SetStereoMode(STEREO_MODE mode)
        {
            stereoMode = mode;
        }

        private void UpdateShaderProperties()
        {
            if (stereoSBSMaterial != null)
            {
                stereoSBSMaterial.SetFloat("_SourceX", sourceX);
                stereoSBSMaterial.SetFloat("_SourceY", sourceY);
                stereoSBSMaterial.SetFloat("_SourceWidth", sourceWidth);
                stereoSBSMaterial.SetFloat("_SourceHeight", sourceHeight);
            }
        }
    }

}
