using System.Collections;
using System.IO;
using UnityEngine;
using UnityEngine.SceneManagement;
#if UNITY_STANDALONE_WIN && KLAK_SPOUT
using Klak.Spout;
#endif
#if KLAK_NDI
using Klak.Ndi;
#endif
#if IGLOO_UI
using Igloo.UI;
#endif


namespace Igloo.Common
{
#pragma warning disable IDE0090 // Use New()...
#pragma warning disable IDE0051 // Remove unused private members

    /// <summary>
    /// The Igloo Manager Class
    /// </summary>
    /// <remarks>
    /// This class controls, sets up, and handles, most of the Igloo camera systems large functions.
    /// It also serves as a reference for the sub classes within the Igloo camera system hierarchy. 
    /// </remarks>
    [System.Serializable]
    public class IglooManager : Singleton<IglooManager>
    {
        /// <summary>
        /// The XML Serializer
        /// </summary>
        /// <remarks>
        /// Turns the XML script into a machine readable format.
        /// </remarks>
        private Serializer serializer;

        /// <summary>
        /// The igloo settings class reference.
        /// </summary>
        [HideInInspector] public Settings settings = null;

        /// <summary>
        /// Creates an Igloo Camera rig, when the IglooManager is established as a Singleton
        /// </summary>
        /// 
        [Tooltip("Creates the camera rig as soon as the start command is called. Can be stopped so the rig is spawned when you call it instead")]
        public bool createOnAwake = true;

        /// <summary>
        /// If true, the IglooManager will not be deleted when the scene is changed.
        /// </summary>
        /// 
        [Tooltip("Keeps the camera rig alive when changing scenes")]
        public bool dontDestroyOnLoad = false;

        /// <summary>
        /// The Igloo Game Object. This is the parent object of the Igloo Camera System
        /// </summary>
        private GameObject igloo;

        /// <summary>
        /// The Camera Prefab GameObject.
        /// </summary>
        /// <remarks>
        /// If a camera prefab is placed here, the IglooManager will instruct the IglooDisplayManager to use 
        /// that camera as the base camera for it's camera system construction. This is useful for adding 
        /// post processing effects to the igloo camera system.
        /// </remarks>
        [Tooltip("Uses this camera as the basis to create the Igloo camera system")]
        public GameObject customCameraPrefab;

        /// <summary>
        /// The Object for the Igloo Camera System to follow.
        /// </summary>
        /// <remarks>
        /// If an object is placed here, the Igloo Player System will be disabled.
        /// Instead the Igloo Follow Object system will be enabled, and will follow the placed object.
        /// </remarks>
        [Tooltip("The object you wish the camera rig to follow. Igloo Camera rig will follow the MainCamera in the scene by Default. Unless Igloo player is installed in which case it will follow that.")]
        public GameObject followObject;

        /// <summary>
        /// The parent object of the Igloo Manager.
        /// </summary>
        [HideInInspector] public GameObject parentObject;

        /// <summary>
        /// The Igloo Network Manager class reference
        /// </summary>
        public NetworkManagerOSC NetworkManager { get; set; }

#if IGLOO_PLAYER
        /// <summary>
        /// The Igloo Player Manager class reference 
        /// </summary>
        public PlayerManager PlayerManager { get; set; }
#endif

        /// <summary>
        /// The Igloo Display Manager class reference
        /// </summary>
        public CameraManager CameraManager { get; set; }

        /// <summary>
        /// The Igloo Head Manager class reference
        /// </summary>
        public HeadManager HeadManager { get; set; }

#if IGLOO_UI
        /// <summary>
        /// The UI Canvas to be used by the Igloo UI System
        /// </summary>
        [Tooltip("The Unity Canvas that should be displayed on the Igloo UI system")]
        public Canvas canvasUI;
#endif

#if KLAK_SPOUT
        /// <summary>
        /// Allows the Spout Sender component to find and use the Spout blit material and shader
        /// which are not included in the build unless this field is allocated. 
        /// If you do not get a spout output after compiling a build, this will likely be the reason. 
        /// The Log will also be filled with null reference errors.
        /// </summary>
        [Tooltip("This is a required file for the Igloo Manager system to send a stream to the Igloo IMP Software. It's located in the Klak Spout package. Click the Add Spout Resources button to locate it")]
        public SpoutResources spoutResources;
#endif

#if KLAK_NDI
        /// <summary>
        /// Allows the NDI Sender component to find and use the NDI blit material and shader
        /// which are not included in the build unless this field is allocated. 
        /// If you do not get a spout output after compiling a build, this will likely be the reason. 
        /// The Log will also be filled with null reference errors.
        /// </summary>
        [Tooltip("This is a required file for the Igloo Manager system to send a stream to the Igloo IMP Software. It's located in the Klak NDI package. Click the Add NDI Resources button to locate it")]
        public NdiResources ndiResources;
#endif


        /// <summary>
        /// Called when the gameobject is initialized as a Singleton
        /// </summary>
        /// <remarks>
        /// Pulls in all the settings from the settings XML, or creates a default set of settings if one cannot 
        /// be found. Also makes the gameobject invenerable, by calling DontDestroyOnLoad.
        /// </remarks>
        protected override void AwakeInternal()
        {
            // Set Igloo Manager to be persistant during scene changes.
            if (Application.isPlaying && dontDestroyOnLoad) DontDestroyOnLoad(this.gameObject);

#if !KLAK_SPOUT
            Debug.LogError("<b>[Igloo]</b> Spout is disabled Apply Project Settings has not been selected. Press Stop, then Click Igloo->Apply Project Settings");
#endif
            // Load XML Settings
            LoadSettings(string.Empty);

            // Load IglooShaders into memory
            StartCoroutine(Utils.LoadShadersFromAssetBundle($@"{Utils.GetICEDataPath()}/iglooshaders"));

            // Setup the Network Manager
            SetupNetworkManager();

            // Initiate the Create Igloo Function
            if (createOnAwake) CreateIgloo();
        }

        /// <summary>
        /// Creates a new Igloo based on the Igloo Settings XML
        /// </summary>
        /// <param name="parent">optional parameter which allows to the Igloo prefab instantiated by CreateIgloo() 
        /// to be parented to the given object. If unassigned the prefab will be parented to the 
        /// IglooManager GameObject. </param>
        /// <param name="follow">optional parameter which assigns the followTransform property</param>
        public void CreateIgloo(GameObject parent = null, GameObject follow = null)
        {
            if (follow) followObject = follow;
            if (parent) parentObject = parent;

            igloo = new GameObject
            {
                name = "IglooCameraRig"
            };
            //igloo.transform.parent = parent == null ? this.gameObject.transform : parent.transform;
            //igloo.transform.localEulerAngles = Vector3.zero;
            //igloo.transform.localPosition = Vector3.zero;

            if (parent)
            {
                igloo.transform.parent = parent == null ? this.gameObject.transform : parent.transform;
                igloo.transform.localEulerAngles = Vector3.zero;
                igloo.transform.localPosition = Vector3.zero;
            }
            else { 
                igloo.transform.position = this.gameObject.transform.position;
                igloo.transform.rotation = this.gameObject.transform.rotation;
            }

            GameObject head = new GameObject
            {
                name = "Head"
            };
            head.transform.parent = igloo.transform;
            head.transform.localEulerAngles = Vector3.zero;
            head.transform.localPosition = Vector3.zero;
            HeadManager = head.AddComponent<HeadManager>();
            CameraManager = igloo.AddComponent<CameraManager>();
            CameraManager.templateCamera = customCameraPrefab;
#if IGLOO_PLAYER
            PlayerManager = igloo.AddComponent<PlayerManager>();
#endif
#if IGLOO_UI
            CreateUIManager();
#endif


            Vector3 tempPos = igloo.transform.localPosition;
            Vector3 tempRot = igloo.transform.localEulerAngles;
            igloo.transform.localEulerAngles = tempRot;
            igloo.transform.localPosition = tempPos;
#if IGLOO_PLAYER
            // player setup
            PlayerManager.GetComponent<CharacterController>().enabled = false;
            PlayerManager.Setup();
            if (settings.PlayerSettings != null && PlayerManager) PlayerManager.SetSettings(settings.PlayerSettings);
#endif
            // Head Setup
            if (settings.HeadSettings != null) HeadManager.SetSettings(settings.HeadSettings);
            FollowObjectTransform followComponent = igloo.AddComponent<FollowObjectTransform>();
            followComponent.enabled = !settings.PlayerSettings.usePlayer;

#if !IGLOO_PLAYER
            followComponent.enabled = true;
            followComponent.followPosition = true;
#endif

            // follow object
            if (followObject)
            {
                followComponent.followObject = followObject;
                followComponent.followPosition = true;
#if IGLOO_PLAYER
                if (PlayerManager) PlayerManager.UsePlayer = false;
                Debug.Log("Igloo - follow trasform assigned, igloo player system disabled");
#endif
            }

#if IGLOO_UI
            SetupUISystem();


            // system settings
            if (settings.SystemSettings != null) SetSystemSettings(settings.SystemSettings);
#endif
        }
#if IGLOO_UI
        private UIManager CreateUIManager()
        {
            GameObject uimgr = new GameObject("UISystem");
            uimgr.transform.SetParent(PlayerManager.transform, false);
            uimgr.transform.SetPositionAndRotation(settings.HeadSettings.headPositionOffset.Vector3, Quaternion.identity);
            return uimgr.AddComponent<UIManager>();
        }

        /// <summary>
        /// Pulled the UI setup out of the start sequence so it can be re-called from other scripts
        /// </summary>
        public void SetupUISystem() {
            if (settings.UISettings.useUI) {
                if (!canvasUI) { Debug.LogWarning("<b>[Igloo]</b> Canvas UI needs to be assigned to the Igloo Manager to enable the UI System."); return; }
                UIManager.instance.SetSettings(settings.UISettings);
                UIManager.instance.canvas = canvasUI;
                UIManager.instance.cursor = FindFirstObjectByType<VirtualInputCursor>().GetComponent<RectTransform>();
                UIManager.instance.texture = Resources.Load<RenderTexture>("RenderTextures/CamTex - Standard");
                canvasUI.renderMode = RenderMode.ScreenSpaceCamera;
                StartCoroutine(DelayedScreenModeSwitch(true));
                UIManager.instance.Setup();
            } 
            else {
                // Not using UI, so switch UI to standard mode.
                Debug.Log("<b>[Igloo]</b> UI System installed but not used.");
            }
        }


        /// <summary>
        /// Waits for one second before calling the Igloo Canvas Controller, as it's not yet created in the main loop.
        /// Stops a null reference error on startup. 
        /// </summary>
        /// <param name="IglooModeOn">If True, calls SwitchToIglooMode on the Canvas Controller</param>
        /// <returns>Null</returns>
        private IEnumerator DelayedScreenModeSwitch(bool IglooModeOn)
        {
            yield return new WaitForSeconds(0.1f);
            if (IglooCanvasController.instance != null)
            {
                if (IglooModeOn)
                {
                    IglooCanvasController.instance.SwitchToIglooMode();
                    canvasUI.worldCamera = IglooCanvasController.instance.vim.m_canvasCamera;
                }
                else IglooCanvasController.instance.SwitchToScreenMode();
            }

        }
#endif
        /// <summary>
        /// Removes the Igloo prefab from the scene if it exists
        /// </summary>
        /// <param name="save"></param>
        public void RemoveIgloo(bool save = true)
        {
            if (igloo)
            {
                if (save) GetAndSaveSettings();
                DestroyImmediate(igloo);
            }
        }

        /// <summary>
        /// Sets up the NetworkManagerOSC system as it's own gameObject
        /// This object will not be destroyed, and is only created if it doesn't already exist.
        /// </summary>
        private void SetupNetworkManager()
        {
            NetworkManager = GetComponent<NetworkManagerOSC>();
            if (settings != null) NetworkManager.SetSettings(settings.NetworkSettings);
            NetworkManager.Setup();
        }

        /// <summary>
        /// Saves settings to the IglooSettings.xml file
        /// These are the current active settings in the scene, not any changed by the settings
        /// editor UI.
        /// </summary>
        public void GetAndSaveSettings()
        {
            var newSettings = new Settings { version = Utils.GetVersion() };
#if IGLOO_PLAYER
            if (PlayerManager) newSettings.PlayerSettings = PlayerManager.GetSettings();
#endif
#if IGLOO_UI
            if (UIManager.instance) newSettings.UISettings = UIManager.instance.GetSettings();
#endif
            if (NetworkManager) newSettings.NetworkSettings = NetworkManager.GetSettings();
            if (HeadManager) newSettings.HeadSettings = HeadManager.GetSettings();
            serializer ??= new Serializer();
            string path = System.IO.Path.Combine(Utils.GetDataPath() + "/", "IglooSettings");
            if (!Directory.Exists(Utils.GetDataPath())) Directory.CreateDirectory(Utils.GetDataPath());
            serializer.Save(path, newSettings);
        }


        /// <summary>
        /// Loads all the XML settings for the creation of the Igloo Camera System. 
        /// if no string argument for <para name="settingsPath"></para> it will load the settings from
        /// the default path of StreamingAssets.
        /// </summary>
        public void LoadSettings(string settingsPath)
        {
            // Pull the Settings from the Settings XML
            serializer ??= new Serializer();
            settings = null;
            string path;
            // if there is a null string in settingsPath, use the standard location of StreamingAssets
            if (settingsPath == string.Empty)
                path = System.IO.Path.Combine(Utils.GetDataPath() + "/", "IglooSettings");
            else path = settingsPath;
            // Load the settings
            settings = serializer.Load(path);

            // if settings cannot be loaded , setup using the default cylinder configuration
            if (settings == null)
            {
                Debug.LogWarning("<b>[Igloo]</b> unable to load Settings XML, using default instead");
                settings = DefaultConfigurations.Shipping();
            }
            else Debug.Log("<b>[Igloo]</b> Settings XML loaded successfully");
        }

        /// <summary>
        /// If a setting has been changed that requires a soft reset.
        /// This function will destroy the current Igloo Camera System. 
        /// Load the new settings XML, and then restart the igloo camera system with the new settings.
        /// </summary>
        public void SoftReset()
        {
            RemoveIgloo(true);
            // Awake internal is currently called, as it will load all settings, then create the Igloo.
            AwakeInternal();
        }

        /// <summary>
        /// Hard reset system, reloads the current scene.
        /// May cause issues if there is a scene system that contains multiple scenes loaded at runtime. 
        /// A better alternative might require quitting the application and reloading it. 
        /// </summary>
        public void HardReset()
        {
            SceneManager.LoadSceneAsync(SceneManager.GetActiveScene().buildIndex);
        }

        /// <summary>
        /// A delayed Coroutine to send a startup message to the Igloo Warper System.
        /// </summary>
        /// <returns>Null</returns>
        IEnumerator SendStartupMessage()
        {
            // Wait for camera setup to complete
            yield return new WaitForSeconds(1.5f);
            if (NetworkManager)
            {
                NetworkMessage msg = new NetworkMessage
                {
                    address = "/externalApplication/selected/enabled"
                };
                msg.AddBoolArgument(true);
                NetworkManager.SendMessage(msg);
            }
        }

        /// <summary>
        /// If the Igloo Manager is destoryed, it will attempt to save the IglooSettings if required.
        /// </summary>
        private void OnDestroy()
        {
            Debug.Log("<b>[Igloo]</b> Manager destructor called");
            if (igloo) Destroy(igloo);
        }


    }
    public class SettingsUI
    {
        public enum SettingsType { Project, File }
        public SettingsType settingsType;

        public enum ConfigType { Development, Shipping }
        public ConfigType configType = ConfigType.Development;


        Vector2 scrollPos;

        public Settings LoadDefaultSettings(ConfigType type = ConfigType.Development)
        {
            return DefaultConfigurations.Shipping();
        }
    }
}

