using UnityEngine;
using System.Collections.Generic;

namespace Igloo
{
    public class OffAxisCamera : IglooCamera
    {
        public RenderTexture effectsChainInputTexture;

        public GameObject head;

        [System.Serializable]
        public class CameraConfig
        {
            public Camera camera;
            private GameObject lowerLeft_;
            private GameObject lowerRight_;
            private GameObject upperLeft_;
            public Vector4 roi;

            public Vector3 LowerLeft => lowerLeft_.transform.position;
            public Vector3 LowerRight => lowerRight_.transform.position;
            public Vector3 UpperLeft => upperLeft_.transform.position;

            public CameraConfig(GameObject parent, Camera cam, IglooCameraData.OffAxisCameraData data)
            {
                GameObject screen = new GameObject($"Viewport - {data.id}");
                screen.transform.parent = parent.transform;
                screen.transform.localPosition = Vector3.zero;

                lowerLeft_ = new GameObject("Lower Left");
                lowerLeft_.transform.parent = screen.transform;
                lowerLeft_.transform.localPosition = new Vector3(data.lowerLeft.x, data.lowerLeft.y, data.lowerLeft.z);

                lowerRight_ = new GameObject("Lower Right");
                lowerRight_.transform.parent = screen.transform;
                lowerRight_.transform.localPosition = new Vector3(data.lowerRight.x, data.lowerRight.y, data.lowerRight.z);

                upperLeft_ = new GameObject("Upper Left");
                upperLeft_.transform.parent = screen.transform;
                upperLeft_.transform.localPosition = new Vector3(data.upperLeft.x, data.upperLeft.y, data.upperLeft.z);

                camera = cam;
                roi = data.roi;
            }
        }

        public List<CameraConfig> cameras = new List<CameraConfig>();

        public override void Setup(IglooCameraData data, GameObject templateCameraGameObject = null)
        {
            base.Setup(data, templateCameraGameObject);

            head = new GameObject("Head");
            head.transform.parent = this.transform;
            head.transform.localPosition = Vector3.zero;

            useWarpAndBlend = data.useWarpAndBlend;

            IglooCameraData.OffAxisCamerasData offAxisCamerasData = data.OffAxisData;

            if (useWarpAndBlend)
            {
                effectsChainInputTexture = new RenderTexture(offAxisCamerasData.CanvasWidth, offAxisCamerasData.CanvasHeight, 24);
                effectsChain = gameObject.AddComponent<EffectsChain>();
                WarpBlendEffect warpBlendEffect = new WarpBlendEffect();
                effectsChain.AddEffect(warpBlendEffect);
            }
            else {
                finalTexture = new RenderTexture(offAxisCamerasData.CanvasWidth, offAxisCamerasData.CanvasHeight, 24);
            }

            head.transform.localPosition = Vector3.zero;

            foreach (var cameraData in offAxisCamerasData.cameras)
            {
                GameObject eyes = new GameObject($"Eye Center - {cameraData.id}");

                GameObject cameraGO;
                if (templateCameraGameObject != null)
                {
                    cameraGO = Instantiate(templateCameraGameObject, this.transform);
                    cameraGO.name = $"Camera_{cameraData.id}";
                }
                else
                {
                    cameraGO = new GameObject($"Camera - {cameraData.id}");
                }

                Camera newCamera = cameraGO.GetComponent<Camera>() ? cameraGO.GetComponent<Camera>() : cameraGO.AddComponent<Camera>();
                if (cameraGO.GetComponent<AudioListener>()) Destroy(cameraGO.GetComponent<AudioListener>());

                CameraConfig config = new CameraConfig(this.gameObject, newCamera, cameraData);

                
                newCamera.fieldOfView = 90;
                newCamera.nearClipPlane = 0.1f;
                newCamera.farClipPlane = 1000f;

                Vector3 dir1 = cameraData.lowerRight - cameraData.lowerLeft;
                Vector3 dir2 = cameraData.upperLeft - cameraData.lowerLeft;
                Vector3 normal = Vector3.Cross(dir1, dir2).normalized;             
                // normal = -normal;
               
                eyes.transform.rotation = Quaternion.LookRotation(normal, dir2.normalized);
                eyes.transform.parent = head.transform;
                eyes.transform.localPosition = Vector3.zero;

                newCamera.transform.parent = eyes.transform;
                newCamera.transform.localEulerAngles = Vector3.zero;
                 
                newCamera.enabled = true;

                newCamera.targetTexture = useWarpAndBlend ? effectsChainInputTexture : finalTexture;
                newCamera.rect = new Rect(config.roi.x, config.roi.y, config.roi.w, config.roi.z);

#if ENABLE_NVIDIA
                // newCamera.GetComponent<UnityEngine.Rendering.HighDefenition.HDAdditionalCameraData>() = true;
#endif
                cameras.Add(config);


            }

            if (useWarpAndBlend) finalTexture = effectsChain.ApplyEffects(effectsChainInputTexture);

            this.transform.localEulerAngles = data.cameraRotationOffset;
        }

        private void Update()
        {
            foreach (var config in cameras)
            {
                if (config.camera != null)
                {
                    config.camera.projectionMatrix = Utils.GetAsymProjMatrix(config.LowerLeft, config.LowerRight, config.UpperLeft, config.camera.gameObject.transform.position, 0.01f, 1000f);
                }
            }

            SetIPD(ipd);

            if (useWarpAndBlend) finalTexture = effectsChain.ApplyEffects(effectsChainInputTexture);
        }

        public override void SetIPD(float val)
        {
            ipd = val;
            
            foreach (var config in cameras)
            {
                if (config.camera != null)
                {
                    config.camera.transform.localPosition = new Vector3(ipd,0,0);
                }
            }
        }


        void OnDrawGizmos()
        {
            foreach (var config in cameras)
            {
                Gizmos.color = Color.blue;

                Vector3 upperRight = config.LowerRight + (config.UpperLeft - config.LowerLeft);

                Vector3 gizmoLowerLeft = config.LowerLeft;
                Vector3 gizmoLowerRight = config.LowerRight;
                Vector3 gizmoUpperLeft = config.UpperLeft;
                Vector3 gizmoUpperRight = upperRight;

                Gizmos.DrawLine(gizmoLowerLeft, gizmoLowerRight);
                Gizmos.DrawLine(gizmoLowerRight, gizmoUpperRight);
                Gizmos.DrawLine(gizmoUpperRight, gizmoUpperLeft);
                Gizmos.DrawLine(gizmoUpperLeft, gizmoLowerLeft);

                Gizmos.color = Color.green;

                Vector3 cameraPosition = config.camera.transform.position;
                Gizmos.DrawLine(cameraPosition, gizmoLowerLeft);
                Gizmos.DrawLine(cameraPosition, gizmoLowerRight);
                Gizmos.DrawLine(cameraPosition, gizmoUpperLeft);
                Gizmos.DrawLine(cameraPosition, gizmoUpperRight);
            }
        }


    }
}






