using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using UnityEngine;

namespace Igloo
{
    public static class TruePerspectiveXML
    {
        public static XDocument LoadAndWrapXml(string xmlPath)
        {
            string xmlContent = File.ReadAllText(xmlPath);
            string wrappedXmlContent = $"<root>{xmlContent}</root>";
            XDocument doc = XDocument.Parse(wrappedXmlContent);
            return doc;
        }


        private static Vector3 ParseVector3(XElement element)
        {
            return new Vector3(
                float.Parse(element.Element("X").Value, CultureInfo.InvariantCulture),
                float.Parse(element.Element("Y").Value, CultureInfo.InvariantCulture),
                float.Parse(element.Element("Z").Value, CultureInfo.InvariantCulture)
            );
        }

        public class CameraConfig
        {
            public Material material_;
            public RenderTexture output_;
            public string Type { get; set; }
            public Rect ROI { get; set; }

            public virtual void LoadMaterial()
            {

            }
        }

        public class CylinderConfig : CameraConfig
        {
            public float Height { get; set; }
            public float Radius { get; set; }
            public float ThetaRange { get; set; }
            public float ThetaOffset { get; set; }

            public override void LoadMaterial()
            {
                //material_ = new Material(Utils.IglooShaders["Igloo/TrueCylinder"]);
                material_ = new Material(Shader.Find("Igloo/TrueCylinder"));

            }
        }

        public class DiskConfig : CameraConfig
        {
            public Vector3 PA { get; set; }
            public Vector3 PB { get; set; }
            public Vector3 PC { get; set; }
            public float Radius { get; set; }
            public float ThetaRange { get; set; }
            public float ThetaOffset { get; set; }

            public Material scaleMaterial;
            public RenderTexture scaledOutputTexture;

            public override void LoadMaterial()
            {
                //material_ = new Material(Utils.IglooShaders["Igloo/TrueDisk"]);
                material_ = new Material(Shader.Find("Igloo/TrueDisk"));
                scaleMaterial = new Material(Shader.Find("Igloo/Scale"));
                scaledOutputTexture = new RenderTexture(output_.width, output_.height, 0);
                // TODO - Test with ThetaRange values other than 360 and 180 
                scaleMaterial.SetFloat("_ThetaRange", ThetaRange);
            }
        }

        public class PlaneConfig : CameraConfig
        {
            public Vector3 PA { get; set; }
            public Vector3 PB { get; set; }
            public Vector3 PC { get; set; }
        }


        private const string CylinderType = "Cylinder";
        private const string DiskType = "Disk";
        private const string PlaneType = "Plane";

        public static CameraConfig CreateConfig(string type)
        {
            switch (type)
            {
                case CylinderType:
                    return new CylinderConfig();
                case DiskType:
                    return new DiskConfig();
                case PlaneType:
                    return new PlaneConfig();
                default:
                    throw new NotSupportedException($"<b>[Igloo]</b> Camera type is unsupported: {type}");
            }
        }

        public static List<CameraConfig> Load(out int outputWidth, out int outputHeight)
        {
            string xmlPath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), "Igloo Vision\\IglooCoreEngine\\settings\\TruePerspective.xml");
            if (!File.Exists(xmlPath))
            {
                Debug.LogWarning("<b>[Igloo]</b> TruePerspective.xml cannot be found");
                outputHeight = 0; outputWidth = 0;
                return null;
            }
            List<CameraConfig> cameras = new List<CameraConfig>();

            XDocument doc = LoadAndWrapXml(xmlPath);

            outputWidth = int.Parse(doc.Descendants("width").First().Value);
            outputHeight = int.Parse(doc.Descendants("height").First().Value);

            foreach (var cameraElem in doc.Descendants("camera"))
            {
                var type = cameraElem.Element("type").Value;
                CameraConfig config = CreateConfig(type);

                var roiElem = cameraElem.Element("roi");
                config.ROI = new Rect(
                    float.Parse(roiElem.Element("X").Value, CultureInfo.InvariantCulture),
                    float.Parse(roiElem.Element("Y").Value, CultureInfo.InvariantCulture),
                    float.Parse(roiElem.Element("W").Value, CultureInfo.InvariantCulture),
                    float.Parse(roiElem.Element("H").Value, CultureInfo.InvariantCulture)
                );

                if (config is CylinderConfig cylinderConfig)
                {
                    cylinderConfig.Height = float.Parse(cameraElem.Element("height").Value, CultureInfo.InvariantCulture);
                    cylinderConfig.Radius = float.Parse(cameraElem.Element("radius").Value, CultureInfo.InvariantCulture);
                    cylinderConfig.ThetaRange = float.Parse(cameraElem.Element("theta_range").Value, CultureInfo.InvariantCulture);
                    cylinderConfig.ThetaOffset = float.Parse(cameraElem.Element("theta_offset").Value, CultureInfo.InvariantCulture);
                }
                else if (config is DiskConfig diskConfig)
                {
                    diskConfig.PA = ParseVector3(cameraElem.Element("pa"));
                    diskConfig.PB = ParseVector3(cameraElem.Element("pb"));
                    diskConfig.PC = ParseVector3(cameraElem.Element("pc"));
                    diskConfig.Radius = float.Parse(cameraElem.Element("radius").Value, CultureInfo.InvariantCulture);
                    diskConfig.ThetaRange = float.Parse(cameraElem.Element("theta_range").Value, CultureInfo.InvariantCulture);
                    diskConfig.ThetaOffset = float.Parse(cameraElem.Element("theta_offset").Value, CultureInfo.InvariantCulture);
                }
                else if (config is PlaneConfig planeConfig)
                {
                    planeConfig.PA = ParseVector3(cameraElem.Element("pa"));
                    planeConfig.PB = ParseVector3(cameraElem.Element("pb"));
                    planeConfig.PC = ParseVector3(cameraElem.Element("pc"));
                }
                cameras.Add(config);
            }
            return cameras;
        }
    }
}
